# -*- test-case-name: vumi.transports.api.tests.test_api -*-
import json

from twisted.python import log
from twisted.internet.defer import inlineCallbacks

from vumi.transports.httprpc import HttpRpcTransport
from vumi.config import ConfigBool, ConfigList, ConfigDict


class HttpApiConfig(HttpRpcTransport.CONFIG_CLASS):
    "HTTP API configuration."
    reply_expected = ConfigBool(
        "True if a reply message is expected.", default=False, static=True)
    allowed_fields = ConfigList(
        "The list of fields a request is allowed to contain. Defaults to the"
        " DEFAULT_ALLOWED_FIELDS class attribute.", static=True)
    field_defaults = ConfigDict(
        "Default values for fields not sent by the client.",
        default={}, static=True)


class HttpApiTransport(HttpRpcTransport):
    """
    Native HTTP API for getting messages into vumi.

    NOTE: This has no security. Put it behind a firewall or something.

    If reply_expected is True, the transport will wait for a reply message
    and will return the reply's content as the HTTP response body. If False,
    the message_id of the dispatched incoming message will be returned.
    """

    transport_type = 'http_api'

    ENCODING = 'utf-8'
    CONFIG_CLASS = HttpApiConfig
    DEFAULT_ALLOWED_FIELDS = (
        'content',
        'to_addr',
        'from_addr',
        'group',
        'session_event',
        )

    def setup_transport(self):
        config = self.get_static_config()
        self.reply_expected = config.reply_expected
        allowed_fields = config.allowed_fields
        if allowed_fields is None:
            allowed_fields = self.DEFAULT_ALLOWED_FIELDS
        self.allowed_fields = set(allowed_fields)
        self.field_defaults = config.field_defaults
        return super(HttpApiTransport, self).setup_transport()

    def handle_outbound_message(self, message):
        if self.reply_expected:
            return super(HttpApiTransport, self).handle_outbound_message(
                message)
        log.msg("HttpApiTransport dropping outbound message: %s" % (message))

    def get_api_field_values(self, request, required_fields):
        values = self.field_defaults.copy()
        errors = {}
        for field in request.args:
            if field not in self.allowed_fields:
                errors.setdefault('unexpected_parameter', []).append(field)
            else:
                values[field] = (
                    request.args.get(field)[0].decode(self.ENCODING))
        for field in required_fields:
            if field not in values and field in self.allowed_fields:
                errors.setdefault('missing_parameter', []).append(field)
        return values, errors

    @inlineCallbacks
    def handle_raw_inbound_message(self, message_id, request):
        values, errors = self.get_api_field_values(request, ['content',
                                                    'to_addr', 'from_addr'])
        if errors:
            yield self.finish_request(message_id, json.dumps(errors), code=400)
            return
        log.msg(('HttpApiTransport sending from %(from_addr)s to %(to_addr)s '
                 'message "%(content)s"') % values)
        payload = {
            'message_id': message_id,
            'transport_type': self.transport_type,
            }
        payload.update(values)
        yield self.publish_message(**payload)
        if not self.reply_expected:
            yield self.finish_request(message_id,
                                      json.dumps({'message_id': message_id}))



from urllib import urlencode

from twisted.internet.defer import inlineCallbacks, returnValue

from zope.interface import implements

from vumi.errors import VumiError
from vumi.tests.helpers import IHelper, generate_proxies
from vumi.transports.tests.helpers import TransportHelper
from vumi.utils import http_request_full


class HttpRpcTransportHelperError(VumiError):
    """Error raised when the HttpRpcTransportHelper encouters an error."""


class HttpRpcTransportHelper(object):
    """
    Test helper for subclasses of
    :class:`~vumi.transports.httprpc.HttpRpcTransport`.

    Adds support for making HTTP requests to the HTTP RPC transport to the
    base :class:`~vumi.transports.tests.helpers.TransportHelper`.

    :param dict request_defaults: Default URL parameters for HTTP requests.

    Other parameters are the same as for
    :class:`~vumi.transports.tests.helpers.TransportHelper`.
    """

    implements(IHelper)

    def __init__(self, transport_class, use_riak=False, request_defaults=None,
                 **msg_helper_args):
        self._transport_helper = TransportHelper(
            transport_class, use_riak=use_riak, **msg_helper_args)

        if request_defaults is None:
            request_defaults = {}
        self.request_defaults = request_defaults
        self.transport_url = None

        generate_proxies(self, self._transport_helper)

    def setup(self, *args, **kw):
        return self._transport_helper.setup(*args, **kw)

    def cleanup(self):
        return self._transport_helper.cleanup()

    @inlineCallbacks
    def get_transport(self, config, cls=None, start=True):
        transport = yield self._transport_helper.get_transport(
            config, cls=cls, start=True)
        self.transport_url = transport.get_transport_url(config['web_path'])
        returnValue(transport)

    def mk_request_raw(self, suffix='', params=None, data=None, method='GET'):
        """
        Make an HTTP request, ignoring this helper's ``request_defaults``.

        :param str suffix: Suffix to add to the transport's URL.
        :param dict params: A dictionary of URL parameters to append to the URL
                            as a query string or None for no URL parameters.
        :param str data: Request body or None for no request body.
        :param str method: HTTP method to use for the request.

        :raises HttpRpcTransportHelperError:
            When invoked before calling :meth:`get_transport`.
        """
        if self.transport_url is None:
            raise HttpRpcTransportHelperError(
                "call .get_transport() before making HTTP requests.")
        url = self.transport_url + suffix
        if params is not None:
            url += '?%s' % urlencode(params)
        return http_request_full(url, data=data, method=method)

    def mk_request(self, _suffix='', _data=None, _method='GET', **kw):
        """
        Make an HTTP request.

        :param str _suffix: Suffix to add to the transport's URL.
        :param str _data: Request body or None for no request body.
        :param str _method: HTTP method to use for the request.
        :param \*\*kw: URL query string parameters.

        :raises HttpRpcTransportHelperError:
            When invoked before calling :meth:`get_transport`.

        The ``_`` prefixes on the function parameter names are to
        make accidental clashes with URL query parameter names less
        likely.
        """
        params = self.request_defaults.copy()
        params.update(kw)
        return self.mk_request_raw(
            suffix=_suffix, params=params, data=_data, method=_method)



# -*- test-case-name: vumi.tests.test_worker -*-

"""Basic tools for workers that handle TransportMessages."""

import time
import os
import socket

from twisted.internet.defer import (
    inlineCallbacks, succeed, maybeDeferred, gatherResults)

from vumi.log import WrappingLogger
from vumi.service import Worker
from vumi.middleware import setup_middlewares_from_config
from vumi.connectors import (
    ReceiveInboundConnector, ReceiveOutboundConnector,
    PublishStatusConnector, ReceiveStatusConnector)
from vumi.config import Config, ConfigInt
from vumi.errors import DuplicateConnectorError
from vumi.utils import generate_worker_id
from vumi.blinkenlights.heartbeat import (HeartBeatPublisher,
                                          HeartBeatMessage)


def then_call(d, func, *args, **kw):
    return d.addCallback(lambda r: func(*args, **kw))


class BaseConfig(Config):
    """Base config definition for workers.

    You should subclass this and add worker-specific fields.
    """

    amqp_prefetch_count = ConfigInt(
        "The number of messages fetched concurrently from each AMQP queue"
        " by each worker instance.",
        default=20, static=True)


class BaseWorker(Worker):
    """Base class for a message processing worker.

    This contains common functionality used by application, transport and
    dispatcher workers. It should be subclassed by workers that need to
    manage their own connectors.
    """

    CONFIG_CLASS = BaseConfig

    def __init__(self, options, config=None):
        super(BaseWorker, self).__init__(options, config=config)
        self.connectors = {}
        self.middlewares = []
        self._static_config = self.CONFIG_CLASS(self.config, static=True)
        self._hb_pub = None
        self._worker_id = None
        self.log = WrappingLogger(system=self.config.get('worker_name'))

    def startWorker(self):
        self.log.msg(
            'Starting a %s worker with config: %s'
            % (self.__class__.__name__, self.config))
        d = maybeDeferred(self._validate_config)
        then_call(d, self.setup_heartbeat)
        then_call(d, self.setup_middleware)
        then_call(d, self.setup_connectors)
        then_call(d, self.setup_worker)
        return d

    def stopWorker(self):
        self.log.msg('Stopping a %s worker.' % (self.__class__.__name__,))
        d = succeed(None)
        then_call(d, self.teardown_worker)
        then_call(d, self.teardown_connectors)
        then_call(d, self.teardown_middleware)
        then_call(d, self.teardown_heartbeat)
        return d

    def setup_connectors(self):
        raise NotImplementedError()

    @inlineCallbacks
    def setup_heartbeat(self):
        # Disable heartbeats if worker_name is not set. We're
        # currently using it as the primary identifier for a worker
        if 'worker_name' in self.config:
            self._worker_name = self.config.get("worker_name")
            self._system_id = self.options.get("system-id", "global")
            self._worker_id = generate_worker_id(self._system_id,
                                                 self._worker_name)
            self.log.msg(
                "Starting HeartBeat publisher with worker_name=%s"
                % self._worker_name)
            self._hb_pub = yield self.start_publisher(
                HeartBeatPublisher, self._gen_heartbeat_attrs)
        else:
            self.log.msg(
                "HeartBeat publisher disabled. No worker_id field found in "
                "config.")

    def teardown_heartbeat(self):
        if self._hb_pub is not None:
            self._hb_pub.stop()
            self._hb_pub = None

    def _gen_heartbeat_attrs(self):
        # worker_name is guaranteed to be set here, otherwise this func would
        # not have been called
        attrs = {
            'version': HeartBeatMessage.VERSION_20130319,
            'worker_id': self._worker_id,
            'system_id': self._system_id,
            'worker_name': self._worker_name,
            'hostname': socket.gethostname(),
            'timestamp': time.time(),
            'pid': os.getpid(),
        }
        attrs.update(self.custom_heartbeat_attrs())
        return attrs

    def custom_heartbeat_attrs(self):
        """Worker subclasses can override this to add custom attributes"""
        return {}

    def teardown_connectors(self):
        d = succeed(None)
        for connector_name in self.connectors.keys():
            then_call(d, self.teardown_connector, connector_name)
        return d

    def setup_worker(self):
        raise NotImplementedError()

    def teardown_worker(self):
        raise NotImplementedError()

    def setup_middleware(self):
        """Create middlewares from config."""
        d = setup_middlewares_from_config(self, self.config)
        d.addCallback(self.middlewares.extend)
        return d

    def teardown_middleware(self):
        """Teardown middlewares."""
        d = succeed(None)
        for mw in reversed(self.middlewares):
            then_call(d, mw.teardown_middleware)
        return d

    def get_static_config(self):
        """Return static (message independent) configuration."""
        return self._static_config

    def get_config(self, msg, ctxt=None):
        """This should return a message and context specific config object.

        It deliberately returns a deferred even when this isn't strictly
        necessary to ensure that workers will continue to work when per-message
        configuration needs to be fetched from elsewhere.
        """
        return succeed(self.CONFIG_CLASS(self.config))

    def _validate_config(self):
        """Once subclasses call `super().validate_config` properly,
           this method can be removed.
           """
        # TODO: remove this once all uses of validate_config have been fixed.
        self.validate_config()

    def validate_config(self):
        """
        Application-specific config validation happens in here.

        Subclasses may override this method to perform extra config
        validation.
        """
        # TODO: deprecate this in favour of a similar method on
        #       config classes.
        pass

    def setup_connector(self, connector_cls, connector_name, middleware=False):
        if connector_name in self.connectors:
            raise DuplicateConnectorError("Attempt to add duplicate connector"
                                          " with name %r" % (connector_name,))
        prefetch_count = self.get_static_config().amqp_prefetch_count
        middlewares = self.middlewares if middleware else None

        connector = connector_cls(self, connector_name,
                                  prefetch_count=prefetch_count,
                                  middlewares=middlewares)
        self.connectors[connector_name] = connector

        d = connector.setup()
        d.addCallback(lambda r: connector)
        return d

    def teardown_connector(self, connector_name):
        connector = self.connectors.pop(connector_name)
        d = connector.teardown()
        d.addCallback(lambda r: connector)
        return d

    def setup_ri_connector(self, connector_name, middleware=True):
        return self.setup_connector(ReceiveInboundConnector, connector_name,
                                    middleware=middleware)

    def setup_ro_connector(self, connector_name, middleware=True):
        return self.setup_connector(ReceiveOutboundConnector, connector_name,
                                    middleware=middleware)

    def setup_publish_status_connector(self, connector_name, middleware=True):
        return self.setup_connector(PublishStatusConnector, connector_name,
                                    middleware=middleware)

    def setup_receive_status_connector(self, connector_name, middleware=True):
        return self.setup_connector(ReceiveStatusConnector, connector_name,
                                    middleware=middleware)

    def pause_connectors(self):
        return gatherResults([
            connector.pause() for connector in self.connectors.itervalues()])

    def unpause_connectors(self):
        for connector in self.connectors.itervalues():
            connector.unpause()



# -*- encoding: utf-8 -*-

import json
from urllib import urlencode

from twisted.internet.defer import inlineCallbacks

from vumi.utils import http_request, http_request_full
from vumi.tests.helpers import VumiTestCase
from vumi.transports.api import HttpApiTransport
from vumi.transports.tests.helpers import TransportHelper


def config_override(**config):
    def deco(fun):
        fun.config_override = config
        return fun
    return deco


class TestHttpApiTransport(VumiTestCase):

    @inlineCallbacks
    def setUp(self):
        self.config = {
            'web_path': "foo",
            'web_port': 0,
        }
        test_method = getattr(self, self._testMethodName)
        config_override = getattr(test_method, 'config_override', {})
        self.config.update(config_override)
        self.tx_helper = self.add_helper(TransportHelper(HttpApiTransport))

        self.transport = yield self.tx_helper.get_transport(self.config)
        self.transport_url = self.transport.get_transport_url()

    def mkurl(self, content, from_addr=123, to_addr=555, **kw):
        params = {
            'to_addr': to_addr,
            'from_addr': from_addr,
            'content': content,
            }
        params.update(kw)
        return self.mkurl_raw(**params)

    def mkurl_raw(self, **params):
        return '%s%s?%s' % (
            self.transport_url,
            self.config['web_path'],
            urlencode(params)
        )

    @inlineCallbacks
    def test_health(self):
        result = yield http_request(
            self.transport_url + "health", "", method='GET')
        self.assertEqual(json.loads(result), {'pending_requests': 0})

    @inlineCallbacks
    def test_inbound(self):
        url = self.mkurl('hello')
        response = yield http_request(url, '', method='GET')
        [msg] = self.tx_helper.get_dispatched_inbound()
        self.assertEqual(msg['transport_name'], self.tx_helper.transport_name)
        self.assertEqual(msg['to_addr'], "555")
        self.assertEqual(msg['from_addr'], "123")
        self.assertEqual(msg['content'], "hello")
        self.assertEqual(json.loads(response),
                         {'message_id': msg['message_id']})

    @inlineCallbacks
    def test_handle_non_ascii_input(self):
        url = self.mkurl(u"öæł".encode("utf-8"))
        response = yield http_request(url, '', method='GET')
        [msg] = self.tx_helper.get_dispatched_inbound()
        self.assertEqual(msg['transport_name'], self.tx_helper.transport_name)
        self.assertEqual(msg['to_addr'], "555")
        self.assertEqual(msg['from_addr'], "123")
        self.assertEqual(msg['content'], u"öæł")
        self.assertEqual(json.loads(response),
                         {'message_id': msg['message_id']})

    @inlineCallbacks
    @config_override(reply_expected=True)
    def test_inbound_with_reply(self):
        d = http_request(self.mkurl('hello'), '', method='GET')
        [msg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        yield self.tx_helper.make_dispatch_reply(msg, "OK")
        response = yield d
        self.assertEqual(response, 'OK')

    @inlineCallbacks
    def test_good_optional_parameter(self):
        url = self.mkurl('hello', group='#channel')
        response = yield http_request(url, '', method='GET')
        [msg] = self.tx_helper.get_dispatched_inbound()
        self.assertEqual(msg['group'], '#channel')
        self.assertEqual(json.loads(response),
                         {'message_id': msg['message_id']})

    @inlineCallbacks
    def test_bad_parameter(self):
        url = self.mkurl('hello', foo='bar')
        response = yield http_request_full(url, '', method='GET')
        self.assertEqual(400, response.code)
        self.assertEqual(json.loads(response.delivered_body),
                         {'unexpected_parameter': ['foo']})

    @inlineCallbacks
    def test_missing_parameters(self):
        url = self.mkurl_raw(content='hello')
        response = yield http_request_full(url, '', method='GET')
        self.assertEqual(400, response.code)
        self.assertEqual(json.loads(response.delivered_body),
                         {'missing_parameter': ['to_addr', 'from_addr']})

    @inlineCallbacks
    @config_override(field_defaults={'to_addr': '555'})
    def test_default_parameters(self):
        url = self.mkurl_raw(content='hello', from_addr='123')
        response = yield http_request(url, '', method='GET')
        [msg] = self.tx_helper.get_dispatched_inbound()
        self.assertEqual(msg['transport_name'], self.tx_helper.transport_name)
        self.assertEqual(msg['to_addr'], "555")
        self.assertEqual(msg['from_addr'], "123")
        self.assertEqual(msg['content'], "hello")
        self.assertEqual(json.loads(response),
                         {'message_id': msg['message_id']})

    @inlineCallbacks
    @config_override(field_defaults={'to_addr': '555'},
                     allowed_fields=['content', 'from_addr'])
    def test_disallowed_default_parameters(self):
        url = self.mkurl_raw(content='hello', from_addr='123')
        response = yield http_request(url, '', method='GET')
        [msg] = self.tx_helper.get_dispatched_inbound()
        self.assertEqual(msg['transport_name'], self.tx_helper.transport_name)
        self.assertEqual(msg['to_addr'], "555")
        self.assertEqual(msg['from_addr'], "123")
        self.assertEqual(msg['content'], "hello")
        self.assertEqual(json.loads(response),
                         {'message_id': msg['message_id']})

    @inlineCallbacks
    @config_override(allowed_fields=['content', 'from_addr'])
    def test_disallowed_parameters(self):
        url = self.mkurl('hello')
        response = yield http_request_full(url, '', method='GET')
        self.assertEqual(400, response.code)
        self.assertEqual(json.loads(response.delivered_body),
                         {'unexpected_parameter': ['to_addr']})



# -*- coding: utf-8 -*-

"""Utilties for marking up RPC methods."""

import inspect
import textwrap
import functools
import itertools

from twisted.internet.defer import Deferred


class RpcCheckError(Exception):
    """Raised when a value fails a type check."""


class Signature(object):

    NO_DEFAULT = object()
    NO_ARG = object()

    def __init__(self, f, returns=None, requires_self=True, **kw):
        self.returns = returns if returns is not None else Null()
        self.requires_self = requires_self
        self.params = kw
        self.argspec = inspect.getargspec(f)
        self.defaults = [self.NO_DEFAULT] * (
            len(self.argspec.args) - len(self.argspec.defaults or ()))
        self.defaults += list(self.argspec.defaults or ())

    def check_params(self, args, kw):
        if kw:
            raise RpcCheckError("Keyword parameters not yet supported.")
        if len(args) > len(self.argspec.args):
            raise RpcCheckError("Too many positional arguments.")

        missing_arg_count = len(self.argspec.args) - len(args)
        args = list(args) + [self.NO_ARG] * missing_arg_count
        arg_tuples = itertools.izip(self.argspec.args, self.defaults, args)
        if self.requires_self:
            next(arg_tuples)

        for arg_name, default, arg_value in arg_tuples:
            if arg_value is self.NO_ARG:
                arg_value = default
            if arg_value is self.NO_DEFAULT:
                raise RpcCheckError("Positional argument %r missing"
                                    " but no default is available." % arg_name)
            arg_type = self.params[arg_name]
            arg_type.check(arg_name, arg_value)

    def check_result(self, result):
        self.returns.check('return value', result)
        return result

    def _wrap_help(self, help_text):
        indent = '    '
        return textwrap.wrap(help_text, initial_indent=indent,
                             subsequent_indent=indent)

    def _format_param(self, param_name, param_type, default):
        lines = [":param %s %s:" % (param_type.name, param_name)]
        help_text = param_type.help()
        if param_type.nullable():
            help_text += " May be null."
        if default is not self.NO_DEFAULT:
            help_text += " Default: %r." % (default,)
        lines.extend(self._wrap_help(help_text))
        return lines

    def _format_return(self, param_type):
        lines = [":rtype %s:" % (param_type.name,)]
        lines.extend(self._wrap_help(param_type.help()))
        return lines

    def _args_with_defaults(self):
        args_defaults = itertools.izip(self.argspec.args, self.defaults)
        if self.requires_self:
            next(args_defaults)

        for arg, default in args_defaults:
            yield arg, self.params[arg], default

    def param_doc(self):
        lines = []
        for arg, arg_type, default in self._args_with_defaults():
            lines.extend(self._format_param(arg, self.params[arg], default))
        lines.extend(self._format_return(self.returns))
        return lines

    def jsonrpc_signature(self):
        sig = [self.returns.jsonrpc_type]
        sig.extend(arg_type.jsonrpc_type for _, arg_type, _
                   in self._args_with_defaults())
        return [sig]


def signature(**kw):
    def decorator(f):
        sig = Signature(f, **kw)

        def wrapper(*args, **kw):
            sig.check_params(args, kw)
            result = f(*args, **kw)
            if isinstance(result, Deferred):
                result.addCallback(sig.check_result)
            else:
                sig.check_result(result)
            return result

        functools.update_wrapper(wrapper, f)
        doc = textwrap.wrap(wrapper.__doc__ or '')
        doc.append("")
        doc.extend(sig.param_doc())
        wrapper.__doc__ = "\n".join(doc)
        wrapper.signature = sig.jsonrpc_signature()
        wrapper.signature_object = sig
        return wrapper

    return decorator


class RpcType(object):

    # See: http://xmlrpc.scripting.com/spec.html
    # valid simple types are:
    #    int, boolean, string, double, base64 and dateTime.iso8601
    # valid compound types are:
    #    array, struct
    jsonrpc_type = None

    def __init__(self, help=None, null=False):
        self._help = help
        self._null = null

    @property
    def name(self):
        return self.__class__.__name__

    def help(self):
        return self._help or ''

    def nullable(self):
        return self._null

    def check(self, name, value):
        if value is None:
            if not self._null:
                raise RpcCheckError("%s may not be None (got None)" % (name,))
            return
        self.nonnull_check(name, value)

    def nonnull_check(self, name, value):
        raise RpcCheckError("The base class RpcType accepts no values.")


class Null(RpcType):
    jsonrpc_type = 'null'

    def __init__(self, *args, **kw):
        kw.setdefault('null', True)
        super(Null, self).__init__(*args, **kw)

    def nonnull_check(self, name, value):
        if value is not None:
            raise RpcCheckError("Null value expected for %s (got %r)"
                                % (name, value))


class Unicode(RpcType):
    jsonrpc_type = 'string'

    def nonnull_check(self, name, value):
        if not isinstance(value, unicode):
            raise RpcCheckError("Unicode value expected for %s (got %r)"
                                % (name, value))


class Int(RpcType):
    jsonrpc_type = 'int'

    def nonnull_check(self, name, value):
        if not isinstance(value, (int, long)):
            raise RpcCheckError("Int value expected for %s (got %r)"
                                % (name, value))


class List(RpcType):
    jsonrpc_type = 'array'

    def __init__(self, *args, **kw):
        self._item_type = kw.pop('item_type', None)
        self._length = kw.pop('length', None)
        super(List, self).__init__(*args, **kw)

    def nonnull_check(self, name, value):
        if not isinstance(value, list):
            raise RpcCheckError("List value expected for %s (got %r)"
                                % (name, value))
        if self._length is not None and len(value) != self._length:
            raise RpcCheckError("List value for %s expected to have"
                                " length %d (got %r)"
                                % (name, self._length, value))
        if self._item_type is not None:
            item_name = 'items of %s' % (name,)
            for item in value:
                self._item_type.check(item_name, item)


class Dict(RpcType):
    jsonrpc_type = 'struct'

    def __init__(self, *args, **kw):
        self._item_type = kw.pop('item_type', None)
        self._required_fields = kw.pop('required_fields', {})
        self._optional_fields = kw.pop('optional_fields', {})
        self._closed = kw.pop('closed', False)
        self._no_checks = all(not x for x in (
            self._item_type, self._required_fields, self._optional_fields,
            self._closed))
        super(Dict, self).__init__(*args, **kw)

    def nonnull_check(self, name, value):
        if not isinstance(value, dict):
            raise RpcCheckError("Dict value expected for %s (got %r)"
                                % (name, value))
        if self._no_checks:
            return
        for key in value:
            field_type = self._required_fields.get(key)
            field_type = (self._optional_fields.get(key)
                          if field_type is None else field_type)
            if field_type is None:
                if self._closed:
                    raise RpcCheckError("Dict received unexpected key %s"
                                        " (got %r)" % (key, value))
                field_type = self._item_type
            if field_type is not None:
                field_type.check('item %s of %s' % (key, name), value[key])
        for key in self._required_fields:
            if key not in value:
                raise RpcCheckError("Dict requires key %s (got %r)"
                                    % (key, value))


class Tag(RpcType):
    jsonrpc_type = 'array'

    def nonnull_check(self, name, value):
        if not isinstance(value, (list, tuple)):
            raise RpcCheckError("Tag %s must be a list or tuple (got %r)"
                                % (name, value))
        if len(value) != 2:
            raise RpcCheckError("Tag %s must contain two elements, a pool name"
                                " and a tag name (got %r)"
                                % (name, value))
        for item in value:
            if not isinstance(item, unicode):
                raise RpcCheckError("Tag %s must have unicode pool and tag"
                                    " name (got %r)" % (name, value))



# -*- test-case-name: vumi.transports.mtech_ussd.tests.test_mtech_ussd -*-

from xml.etree import ElementTree as ET

from twisted.internet.defer import inlineCallbacks

from vumi import log
from vumi.message import TransportUserMessage
from vumi.transports.httprpc import HttpRpcTransport
from vumi.components.session import SessionManager


class MtechUssdTransport(HttpRpcTransport):
    """MTECH USSD transport.

    Configuration parameters:

    :param str transport_name:
        The name this transport instance will use to create its queues
    :param int ussd_session_timeout:
        Number of seconds before USSD session information stored in
        Redis expires. Default is 600s.
    :param str web_path:
        The HTTP path to listen on.
    :param int web_port:
        The HTTP port to listen on.

    NOTE: We currently only support free-text USSD, not menus.
          At the time of writing, vumi has no suitable message format for
          specifying USSD menus. This may change in the future.
    """
    ENCODING = 'utf-8'

    @inlineCallbacks
    def setup_transport(self):
        super(MtechUssdTransport, self).setup_transport()
        r_config = self.config.get('redis_manager', {})
        r_prefix = "mtech_ussd:%s" % self.transport_name
        session_timeout = int(self.config.get("ussd_session_timeout", 600))
        self.session_manager = yield SessionManager.from_redis_config(
                r_config, r_prefix, max_session_length=session_timeout)

    @inlineCallbacks
    def teardown_transport(self):
        yield self.session_manager.stop()
        yield super(MtechUssdTransport, self).teardown_transport()

    def save_session(self, session_id, from_addr, to_addr):
        return self.session_manager.create_session(
            session_id, from_addr=from_addr, to_addr=to_addr)

    def handle_status_message(self, msgid, session_id):
        mur = MtechUssdResponse(session_id)
        response_body = unicode(mur).encode(self.ENCODING)
        log.msg("Outbound message: %r" % (response_body,))
        return self.finish_request(msgid, response_body)

    @inlineCallbacks
    def handle_raw_inbound_message(self, msgid, request):
        request_body = request.content.read()
        log.msg("Inbound message: %r" % (request_body,))
        try:
            body = ET.fromstring(request_body)
        except:
            log.warning("Error parsing request XML: %s" % (request_body,))
            yield self.finish_request(msgid, "", code=400)
            return

        # We always get this.
        session_id = body.find('session_id').text

        status_elem = body.find('status')
        if status_elem is not None:
            # We have a status message. These are all variations on "cancel".
            yield self.handle_status_message(msgid, session_id)
            return

        page_id = body.find('page_id').text

        # They sometimes send us page_id=0 in the middle of a session.
        if page_id == '0' and body.find('mobile_number') is not None:
            # This is a new session.
            session = yield self.save_session(
                session_id,
                from_addr=body.find('mobile_number').text,
                to_addr=body.find('gate').text)  # ???
            session_event = TransportUserMessage.SESSION_NEW
        else:
            # This is an existing session.
            session = yield self.session_manager.load_session(session_id)
            if 'from_addr' not in session:
                # We have a missing or broken session.
                yield self.finish_request(msgid, "", code=400)
                return
            session_event = TransportUserMessage.SESSION_RESUME

        content = body.find('data').text

        transport_metadata = {'session_id': session_id}
        self.publish_message(
                message_id=msgid,
                content=content,
                to_addr=session['to_addr'],
                from_addr=session['from_addr'],
                session_event=session_event,
                transport_name=self.transport_name,
                transport_type=self.config.get('transport_type'),
                transport_metadata=transport_metadata,
                )

    def handle_outbound_message(self, message):
        in_reply_to = message['in_reply_to']
        session_id = message['transport_metadata'].get('session_id')
        content = message['content']
        if not (in_reply_to and session_id and content):
            return self.publish_nack(user_message_id=message['message_id'],
                sent_message_id=message['message_id'],
                reason='Missing in_reply_to, content or session_id')

        mur = MtechUssdResponse(message['transport_metadata']['session_id'])
        mur.add_text(message['content'])
        if message['session_event'] != TransportUserMessage.SESSION_CLOSE:
            mur.add_freetext_option()
        response_body = unicode(mur).encode(self.ENCODING)
        log.msg("Outbound message: %r" % (response_body,))
        self.finish_request(message['in_reply_to'], response_body)
        return self.publish_ack(user_message_id=message['message_id'],
            sent_message_id=message['message_id'])


class MtechUssdResponse(object):
    def __init__(self, session_id):
        self.session_id = session_id
        self.title = None
        self.text = []
        self.nav = []

    def add_title(self, title):
        self.title = title

    def add_text(self, text):
        self.text.append(text)

    def add_menu_item(self, text, option):
        self.nav.append({
                'text': text,
                'pageId': 'index%s' % (option,),
                'accesskey': option,
                })

    def add_freetext_option(self):
        self.nav.append({'text': None, 'pageId': 'indexX', 'accesskey': '*'})

    def to_xml(self):
        page = ET.fromstring('<page version="2.0" />')
        ET.SubElement(page, "session_id").text = self.session_id

        if self.title is not None:
            ET.SubElement(page, "title").text = self.title

        for text in self.text:
            lines = text.split('\n')
            div = ET.SubElement(page, "div")
            div.text = lines.pop(0)
            for line in lines:
                ET.SubElement(div, "br").tail = line

        if self.nav:
            nav = ET.SubElement(page, "navigation")
            for link in self.nav:
                ET.SubElement(
                    nav, "link", pageId=link['pageId'],
                    accesskey=link['accesskey']).text = link['text']

        # We can't have "\n" in the output at all, it seems.
        return ET.tostring(page, encoding="UTF-8").replace("\n", "")

    def __str__(self):
        return self.to_xml()



from vumi.transports.cellulant.cellulant import (
    CellulantTransport, CellulantError)
from vumi.transports.cellulant.cellulant_sms import CellulantSmsTransport

__all__ = ['CellulantTransport', 'CellulantSmsTransport', 'CellulantError']



"""Tests for vumi.reconnecting_client."""

import random

from twisted.internet import interfaces
from twisted.internet.defer import inlineCallbacks, Deferred, CancelledError
from twisted.internet.protocol import Protocol
from twisted.internet.task import Clock
from twisted.python.failure import Failure
from twisted.trial.unittest import TestCase
from zope.interface import implementer

from vumi.reconnecting_client import ReconnectingClientService
from vumi.tests.utils import LogCatcher


@implementer(interfaces.IStreamClientEndpoint)
class ClientTestEndpoint(object):
    def __init__(self):
        self.connect_called = Deferred()
        self.connected = Deferred()

    def connect(self, factory):
        self.connect_called.callback(factory)
        return self.connected


class DummyProtocol(Protocol):
    pass


@implementer(interfaces.ITransport)
class DummyTransport(object):

    def __init__(self):
        self.lose_connection_called = Deferred()

    def loseConnection(self):
        self.lose_connection_called.callback(None)


class MockRecorder(object):
    def __init__(self, test_case, result=None):
        self._test_case = test_case
        self._calls = []
        self._result = result

    def assertCalledOnce(self, *args, **kw):
        self._test_case.assertEqual(self._calls, [(args, kw)])

    def assertNotCalled(self):
        self._test_case.assertEqual(self._calls, [])

    def __call__(self, *args, **kw):
        self._calls.append((args, kw))
        return self._result


class ReconnectingClientServiceTestCase(TestCase):
    def make_reconnector(self, **kw):
        e = ClientTestEndpoint()
        f = object()
        s = ReconnectingClientService(e, f)
        for key, value in kw.items():
            setattr(s, key, value)
        self.addCleanup(s.stopService)
        return s, e, f

    def patch_reconnector(self, method):
        mock = MockRecorder(self)
        self.patch(ReconnectingClientService, method, mock)
        return mock

    def test_startService(self):
        retry = self.patch_reconnector('retry')
        s = ReconnectingClientService(object(), object())
        s.startService()
        self.assertTrue(s.continueTrying)
        retry.assertCalledOnce(delay=0.0)

    @inlineCallbacks
    def test_stopService(self):
        s, e, f = self.make_reconnector(continueTrying=True)
        yield s.stopService()
        self.assertEqual(s.continueTrying, False)

    @inlineCallbacks
    def test_stopService_while_retrying(self):
        s, e, f = self.make_reconnector()
        clock = Clock()
        r = s._delayedRetry = clock.callLater(1.0, lambda: None)
        yield s.stopService()
        self.assertTrue(r.cancelled)
        self.assertIdentical(s._delayedRetry, None)

    @inlineCallbacks
    def test_stopService_while_connecting(self):
        errs = []
        s, e, f = self.make_reconnector()
        s._connectingDeferred = Deferred().addErrback(errs.append)
        yield s.stopService()
        [failure] = errs
        self.assertTrue(failure.check(CancelledError))

    @inlineCallbacks
    def test_stopService_while_connected(self):
        s, e, f = self.make_reconnector()
        s._protocol = DummyProtocol()
        s._protocol.transport = DummyTransport()
        d = s.stopService()
        self.assertFalse(d.called)
        self.assertTrue(s._protocol.transport.lose_connection_called.called)
        s.clientConnectionLost(Failure(Exception()))
        yield d

    def test_clientConnected(self):
        reset = self.patch_reconnector('resetDelay')
        s = ReconnectingClientService(object(), object())
        p = object()
        s.clientConnected(p)
        self.assertIdentical(s._protocol, p)
        reset.assertCalledOnce()

    def test_clientConnectionFailed(self):
        retry = self.patch_reconnector('retry')
        s = ReconnectingClientService(object(), object())
        s.clientConnectionFailed(Failure(Exception()))
        self.assertIdentical(s._protocol, None)
        retry.assertCalledOnce()

    def test_clientConnectionLost(self):
        retry = self.patch_reconnector('retry')
        s = ReconnectingClientService(object(), object())
        s.clientConnectionLost(Failure(Exception()))
        self.assertIdentical(s._protocol, None)
        retry.assertCalledOnce()

    def test_clientConnectionLost_while_stopping(self):
        retry = self.patch_reconnector('retry')
        s = ReconnectingClientService(object(), object())
        d = s._protocolStoppingDeferred = Deferred()
        s.clientConnectionLost(Failure(Exception()))
        self.assertIdentical(s._protocol, None)
        self.assertIdentical(s._protocolStoppingDeferred, None)
        retry.assertCalledOnce()
        self.assertTrue(d.called)

    def test_retry_aborts_when_stopping(self):
        s, e, f = self.make_reconnector(continueTrying=False)
        s.retry()
        self.assertEqual(s.retries, 0)

    def test_noisy_retry_aborts_when_stopping(self):
        s, e, f = self.make_reconnector(noisy=True, continueTrying=False)
        with LogCatcher() as lc:
            s.retry()
            [msg] = lc.messages()
        self.assertEqual(s.retries, 0)
        self.assertSubstring("Abandoning <vumi.tests.test_reconnecting_client"
                             ".ClientTestEndpoint object at", msg)
        self.assertSubstring("on explicit request", msg)

    def test_retry_aborts_when_max_retries_exceeded(self):
        s, e, f = self.make_reconnector(maxRetries=5, continueTrying=True)
        s.retries = 5
        s.retry()
        self.assertEqual(s.retries, 5)

    def test_noisy_retry_aborts_when_max_retries_exceeded(self):
        s, e, f = self.make_reconnector(noisy=True, maxRetries=5,
                                        continueTrying=True)
        s.retries = 5
        with LogCatcher() as lc:
            s.retry()
            [msg] = lc.messages()
        self.assertEqual(s.retries, 5)
        self.assertSubstring("Abandoning <vumi.tests.test_reconnecting_client"
                             ".ClientTestEndpoint object at", msg)
        self.assertSubstring("after 5 retries", msg)

    def test_retry_with_explicit_delay(self):
        s, e, f = self.make_reconnector(continueTrying=True, clock=Clock())
        s.retry(delay=1.5)
        [delayed] = s.clock.calls
        self.assertEqual(delayed.time, 1.5)

    def test_noisy_retry_with_explicit_delay(self):
        s, e, f = self.make_reconnector(noisy=True, continueTrying=True,
                                        clock=Clock())
        with LogCatcher() as lc:
            s.retry(delay=1.5)
            [msg] = lc.messages()
        [delayed] = s.clock.calls
        self.assertEqual(delayed.time, 1.5)
        self.assertSubstring("Will retry <vumi.tests.test_reconnecting_client"
                             ".ClientTestEndpoint object at", msg)
        self.assertSubstring("in 1.5 seconds", msg)

    def test_retry_delay_advances(self):
        s, e, f = self.make_reconnector(jitter=None, continueTrying=True,
                                        clock=Clock())
        s.retry()
        [delayed] = s.clock.calls
        self.assertAlmostEqual(delayed.time, s.factor)
        self.assertAlmostEqual(s.delay, s.factor)

    def test_retry_delay_is_capped_by_maxDelay(self):
        s, e, f = self.make_reconnector(jitter=None, continueTrying=True,
                                        clock=Clock(), maxDelay=1.5)
        s.retry()
        [delayed] = s.clock.calls
        self.assertAlmostEqual(delayed.time, 1.5)
        self.assertAlmostEqual(s.delay, 1.5)

    def test_retry_with_jitter(self):
        normal = MockRecorder(self, result=2.0)
        self.patch(random, 'normalvariate', normal)
        s, e, f = self.make_reconnector(continueTrying=True, clock=Clock())
        s.retry()
        [delayed] = s.clock.calls
        self.assertAlmostEqual(delayed.time, 2.0)
        self.assertAlmostEqual(s.delay, 2.0)
        normal.assertCalledOnce(s.factor, s.factor * s.jitter)

    @inlineCallbacks
    def test_retry_when_connection_succeeds(self):
        connected = self.patch_reconnector('clientConnected')
        s, e, f = self.make_reconnector(continueTrying=True, clock=Clock())

        s.retry(delay=1.0)
        connected.assertNotCalled()

        s.clock.advance(1.0)
        wrapped_f = yield e.connect_called
        self.assertEqual(wrapped_f.protocolFactory, f)
        connected.assertNotCalled()

        p = DummyProtocol()
        e.connected.callback(p)
        connected.assertCalledOnce(p)

    @inlineCallbacks
    def test_retry_when_connection_fails(self):
        connection_failed = self.patch_reconnector('clientConnectionFailed')
        s, e, f = self.make_reconnector(continueTrying=True, clock=Clock())

        s.retry(delay=1.0)
        connection_failed.assertNotCalled()

        s.clock.advance(1.0)
        wrapped_f = yield e.connect_called
        self.assertEqual(wrapped_f.protocolFactory, f)
        connection_failed.assertNotCalled()

        failure = Failure(Exception())
        e.connected.errback(failure)
        connection_failed.assertCalledOnce(failure)

    def test_resetDelay(self):
        initial_delay = ReconnectingClientService.initialDelay
        s = ReconnectingClientService(object(), object())
        s.delay, s.retries = initial_delay + 1, 5
        s.resetDelay()
        self.assertEqual(s.delay, initial_delay)
        self.assertEqual(s.retries, 0)

    def test_parametrizedClock(self):
        """
        The clock used by L{ReconnectingClientFactory} can be parametrized, so
        that one can cleanly test reconnections.
        """
        clock = Clock()
        s, e, f = self.make_reconnector()
        s.clock = clock
        s.startService()
        self.assertEqual(len(clock.calls), 1)



# -*- test-case-name: vumi.transports.vumi_bridge.tests.test_vumi_bridge -*-

import base64
import json

from twisted.internet.defer import inlineCallbacks
from twisted.web import http
from twisted.web.resource import Resource
from twisted.web.server import NOT_DONE_YET

from vumi.transports import Transport
from vumi.config import ConfigText, ConfigDict, ConfigInt, ConfigFloat
from vumi.persist.txredis_manager import TxRedisManager
from vumi.message import TransportUserMessage, TransportEvent
from vumi.utils import to_kwargs, http_request_full, StatusEdgeDetector
from vumi import log


class VumiBridgeTransportConfig(Transport.CONFIG_CLASS):
    account_key = ConfigText(
        'The account key to connect with.', static=True, required=True)
    conversation_key = ConfigText(
        'The conversation key to use.', static=True, required=True)
    access_token = ConfigText(
        'The access token for the conversation key.', static=True,
        required=True)
    base_url = ConfigText(
        'The base URL for the API', static=True,
        default='https://go.vumi.org/api/v1/go/http_api/')
    message_life_time = ConfigInt(
        'How long to keep message_ids around for.', static=True,
        default=48 * 60 * 60)  # default is 48 hours.
    redis_manager = ConfigDict(
        "Redis client configuration.", default={}, static=True)
    max_reconnect_delay = ConfigInt(
        'Maximum number of seconds between connection attempts', default=3600,
        static=True)
    max_retries = ConfigInt(
        'Maximum number of consecutive unsuccessful connection attempts '
        'after which no further connection attempts will be made. If this is '
        'not explicitly set, no maximum is applied',
        static=True)
    initial_delay = ConfigFloat(
        'Initial delay for first reconnection attempt', default=0.1,
        static=True)
    factor = ConfigFloat(
        'A multiplicitive factor by which the delay grows',
        # (math.e)
        default=2.7182818284590451,
        static=True)
    jitter = ConfigFloat(
        'Percentage of randomness to introduce into the delay length'
        'to prevent stampeding.',
        # molar Planck constant times c, joule meter/mole
        default=0.11962656472,
        static=True)
    web_port = ConfigInt(
        "The port to listen for requests on, defaults to `0`.",
        default=0, static=True)
    web_path = ConfigText(
        "The path to listen for inbound requests on.", required=True,
        static=True)
    health_path = ConfigText(
        "The path to listen for downstream health checks on"
        " (useful with HAProxy)", default='health', static=True)


class GoConversationTransportBase(Transport):

    agent_factory = None  # For swapping out the Agent we use in tests.

    def get_url(self, path):
        config = self.get_static_config()
        url = '/'.join([
            config.base_url.rstrip('/'), config.conversation_key, path])
        return url

    @inlineCallbacks
    def map_message_id(self, remote_message_id, local_message_id):
        config = self.get_static_config()
        yield self.redis.set(remote_message_id, local_message_id)
        yield self.redis.expire(remote_message_id, config.message_life_time)

    def get_message_id(self, remote_message_id):
        return self.redis.get(remote_message_id)

    def handle_inbound_message(self, message):
        return self.publish_message(**message.payload)

    @inlineCallbacks
    def handle_inbound_event(self, event):
        remote_message_id = event['user_message_id']
        local_message_id = yield self.get_message_id(remote_message_id)
        event['user_message_id'] = local_message_id
        event['sent_message_id'] = remote_message_id
        yield self.publish_event(**event.payload)

    @inlineCallbacks
    def handle_outbound_message(self, message):
        headers = {
            'Content-Type': 'application/json; charset=utf-8',
        }
        headers.update(self.get_auth_headers())

        params = {
            'to_addr': message['to_addr'],
            'content': message['content'],
            'message_id': message['message_id'],
            'in_reply_to': message['in_reply_to'],
            'session_event': message['session_event']
        }
        if 'helper_metadata' in message:
            params['helper_metadata'] = message['helper_metadata']

        resp = yield http_request_full(
            self.get_url('messages.json'),
            data=json.dumps(params).encode('utf-8'),
            headers=headers,
            method='PUT',
            agent_class=self.agent_factory)

        if resp.code != http.OK:
            log.warning('Unexpected status code: %s, body: %s' % (
                resp.code, resp.delivered_body))
            self.update_status(
                status='down', component='submitted-to-vumi-go',
                type='bad_request',
                message='Message submission rejected by Vumi Go')
            yield self.publish_nack(message['message_id'],
                                    reason='Unexpected status code: %s' % (
                                        resp.code,))
            return

        remote_message = json.loads(resp.delivered_body)
        yield self.map_message_id(
            remote_message['message_id'], message['message_id'])
        self.update_status(
            status='ok', component='submitted-to-vumi-go',
            type='good_request', message='Message accepted by Vumi Go')
        yield self.publish_ack(user_message_id=message['message_id'],
                               sent_message_id=remote_message['message_id'])

    def get_auth_headers(self):
        config = self.get_static_config()
        return {
            'Authorization': ['Basic ' + base64.b64encode('%s:%s' % (
                config.account_key, config.access_token))],
        }

    @inlineCallbacks
    def update_status(self, **kw):
        '''Publishes a status if it is not a repeat of the previously
        published status.'''
        if self.status_detect.check_status(**kw):
            yield self.publish_status(**kw)


class GoConversationHealthResource(Resource):
    # Most of this copied wholesale from vumi.transports.httprpc.
    isLeaf = True

    def __init__(self, transport):
        self.transport = transport
        Resource.__init__(self)

    def render_GET(self, request):
        request.setResponseCode(http.OK)
        request.do_not_log = True
        return self.transport.get_health_response()


class GoConversationResource(Resource):
    # Most of this copied wholesale from vumi.transports.httprpc.
    isLeaf = True

    def __init__(self, callback):
        self.callback = callback
        Resource.__init__(self)

    def render_(self, request, request_id=None):
        request.setHeader("content-type", 'application/json; charset=utf-8')
        self.callback(request)
        return NOT_DONE_YET

    def render_PUT(self, request):
        return self.render_(request)

    def render_POST(self, request):
        return self.render_(request)


class GoConversationTransport(GoConversationTransportBase):
    # Most of this copied wholesale from vumi.transports.httprpc.

    CONFIG_CLASS = VumiBridgeTransportConfig

    @inlineCallbacks
    def setup_transport(self):
        config = self.get_static_config()
        self.redis = yield TxRedisManager.from_config(
            config.redis_manager)

        self.web_resource = yield self.start_web_resources([
            (GoConversationResource(self.handle_raw_inbound_message),
             "%s/messages.json" % (config.web_path)),
            (GoConversationResource(self.handle_raw_inbound_event),
             "%s/events.json" % (config.web_path)),
            (GoConversationHealthResource(self), config.health_path),
        ], config.web_port)
        self.status_detect = StatusEdgeDetector()

    def teardown_transport(self):
        return self.web_resource.loseConnection()

    def get_transport_url(self, suffix=''):
        """
        Get the URL for the HTTP resource. Requires the worker to be started.

        This is mostly useful in tests, and probably shouldn't be used
        in non-test code, because the API might live behind a load
        balancer or proxy.
        """
        addr = self.web_resource.getHost()
        return "http://%s:%s/%s/%s" % (
            addr.host, addr.port, self.config["web_path"], suffix.lstrip('/'))

    @inlineCallbacks
    def handle_raw_inbound_event(self, request):
        try:
            data = json.loads(request.content.read())
            msg = TransportEvent(_process_fields=True, **to_kwargs(data))
            yield self.handle_inbound_event(msg)
            request.finish()
            if msg.payload["event_type"] == "ack":
                self.update_status(
                    status='ok', component='sent-by-vumi-go',
                    type='vumi_go_sent', message='Sent by Vumi Go')
            elif msg.payload["event_type"] == "nack":
                self.update_status(
                    status='down', component='sent-by-vumi-go',
                    type='vumi_go_failed', message='Vumi Go failed to send')
            self.update_status(
                status='ok', component='vumi-go-event',
                type='good_request',
                message='Good event received from Vumi Go')
        except Exception as e:
            log.err(e)
            request.setResponseCode(400)
            request.finish()
            self.update_status(
                status='down', component='vumi-go-event',
                type='bad_request', message='Bad event received from Vumi Go')

    @inlineCallbacks
    def handle_raw_inbound_message(self, request):
        try:
            data = json.loads(request.content.read())
            msg = TransportUserMessage(
                _process_fields=True, **to_kwargs(data))
            yield self.handle_inbound_message(msg)
            request.finish()
            self.update_status(
                status='ok', component='received-from-vumi-go',
                type='good_request', message='Good request received')
        except Exception as e:
            log.err(e)
            request.setResponseCode(400)
            request.finish()
            self.update_status(
                status='down', component='received-from-vumi-go',
                type='bad_request', message='Bad request received')



"""Tests for vumi.scripts.vumi_redis_tools."""

import StringIO

import yaml

from twisted.python.usage import UsageError

from vumi.scripts.vumi_redis_tools import (
    scan_keys, TaskRunner, Options, Task, TaskError,
    Count, Expire, Persist, ListKeys, Skip)
from vumi.tests.helpers import VumiTestCase, PersistenceHelper


class DummyTaskRunner(object):
    def __init__(self):
        self.output = []

    def emit(self, s):
        self.output.append(s)


class DummyTask(Task):
    """Dummy task for testing."""

    name = "dummy"
    hidden = True

    def __init__(self, a=None, b=None):
        self.a = a
        self.b = b


class TestTask(VumiTestCase):
    def test_name(self):
        t = Task()
        self.assertEqual(t.name, None)

    def test_parse_with_args(self):
        t = Task.parse("dummy:a=foo,b=bar")
        self.assertEqual(t.name, "dummy")
        self.assertEqual(t.a, "foo")
        self.assertEqual(t.b, "bar")
        self.assertEqual(type(t), DummyTask)

    def test_parse_without_args(self):
        t = Task.parse("dummy")
        self.assertEqual(t.name, "dummy")
        self.assertEqual(t.a, None)
        self.assertEqual(t.b, None)
        self.assertEqual(type(t), DummyTask)

    def test_parse_no_task(self):
        self.assertRaises(TaskError, Task.parse, "unknown")

    def test_init(self):
        t = Task()
        runner = object()
        redis = object()
        t.init(runner, redis)
        self.assertEqual(t.runner, runner)
        self.assertEqual(t.redis, redis)


class TestCount(VumiTestCase):

    def setUp(self):
        self.runner = DummyTaskRunner()

    def mk_count(self):
        t = Count()
        t.init(self.runner, None)
        t.before()
        return t

    def test_name(self):
        t = Count()
        self.assertEqual(t.name, "count")

    def test_create(self):
        t = Task.parse("count")
        self.assertEqual(t.name, "count")
        self.assertEqual(type(t), Count)

    def test_before(self):
        t = self.mk_count()
        self.assertEqual(t.count, 0)

    def test_process_key(self):
        t = self.mk_count()
        key = t.process_key("foo")
        self.assertEqual(t.count, 1)
        self.assertEqual(key, "foo")

    def test_after(self):
        t = self.mk_count()
        for i in range(5):
            t.process_key(str(i))
        t.after()
        self.assertEqual(self.runner.output, [
            "Found 5 matching keys.",
        ])


class TestExpire(VumiTestCase):

    def setUp(self):
        self.runner = DummyTaskRunner()
        self.persistence_helper = self.add_helper(
            PersistenceHelper(is_sync=True))
        self.redis = self.persistence_helper.get_redis_manager()
        self.redis._purge_all()  # Make sure we start fresh.

    def mk_expire(self, seconds=10):
        t = Expire(seconds)
        t.init(self.runner, self.redis)
        t.before()
        return t

    def test_name(self):
        t = Expire(seconds=20)
        self.assertEqual(t.name, "expire")

    def test_create(self):
        t = Task.parse("expire:seconds=20")
        self.assertEqual(t.name, "expire")
        self.assertEqual(t.seconds, 20)
        self.assertEqual(type(t), Expire)

    def test_process_key(self):
        t = self.mk_expire(seconds=10)
        self.redis.set("key1", "bar")
        self.redis.set("key2", "baz")
        key = t.process_key("key1")
        self.assertEqual(key, "key1")
        self.assertTrue(
            0 < self.redis.ttl("key1") <= 10)
        self.assertEqual(
            self.redis.ttl("key2"), None)


class TestPersist(VumiTestCase):

    def setUp(self):
        self.runner = DummyTaskRunner()
        self.persistence_helper = self.add_helper(
            PersistenceHelper(is_sync=True))
        self.redis = self.persistence_helper.get_redis_manager()
        self.redis._purge_all()  # Make sure we start fresh.

    def mk_persist(self):
        t = Persist()
        t.init(self.runner, self.redis)
        t.before()
        return t

    def test_name(self):
        t = Persist()
        self.assertEqual(t.name, "persist")

    def test_create(self):
        t = Task.parse("persist")
        self.assertEqual(t.name, "persist")
        self.assertEqual(type(t), Persist)

    def test_process_key(self):
        t = self.mk_persist()
        self.redis.setex("key1", 10, "bar")
        self.redis.setex("key2", 20, "baz")
        key = t.process_key("key1")
        self.assertEqual(key, "key1")
        self.assertEqual(
            self.redis.ttl("key1"), None)
        self.assertTrue(
            0 < self.redis.ttl("key2") <= 20)


class TestListKeys(VumiTestCase):

    def setUp(self):
        self.runner = DummyTaskRunner()

    def mk_list(self):
        t = ListKeys()
        t.init(self.runner, None)
        t.before()
        return t

    def test_name(self):
        t = ListKeys()
        self.assertEqual(t.name, "list")

    def test_create(self):
        t = Task.parse("list")
        self.assertEqual(t.name, "list")
        self.assertEqual(type(t), ListKeys)

    def test_process_key(self):
        t = self.mk_list()
        key = t.process_key("key1")
        self.assertEqual(key, "key1")
        self.assertEqual(self.runner.output, [
            "key1",
        ])


class TestSkip(VumiTestCase):

    def setUp(self):
        self.runner = DummyTaskRunner()

    def mk_skip(self, pattern=".*"):
        t = Skip(pattern)
        t.init(self.runner, None)
        t.before()
        return t

    def test_name(self):
        t = Skip(".*")
        self.assertEqual(t.name, "skip")

    def test_create(self):
        t = Task.parse("skip:pattern=.*")
        self.assertEqual(t.name, "skip")
        self.assertEqual(t.regex.pattern, ".*")
        self.assertEqual(type(t), Skip)

    def test_process_key(self):
        t = self.mk_skip("skip_.*")
        self.assertEqual(t.process_key("skip_this"), None)
        self.assertEqual(t.process_key("dont_skip"), "dont_skip")


class TestOptions(VumiTestCase):
    def mk_file(self, data):
        name = self.mktemp()
        with open(name, "wb") as data_file:
            data_file.write(data)
        return name

    def mk_redis_config(self, key_prefix):
        config = {
            'redis_manager': {
                'key_prefix': key_prefix,
            },
        }
        return self.mk_file(yaml.safe_dump(config))

    def mk_opts_raw(self, args):
        opts = Options()
        opts.parseOptions(args)
        return opts

    def mk_opts(self, args):
        config = self.mk_redis_config("foo")
        return self.mk_opts_raw(args + [config, "*"])

    def test_no_config(self):
        exc = self.assertRaises(
            UsageError,
            self.mk_opts_raw, [])
        self.assertEqual(str(exc), "Wrong number of arguments.")

    def test_no_pattern(self):
        exc = self.assertRaises(
            UsageError,
            self.mk_opts_raw, ["config.yaml"])
        self.assertEqual(str(exc), "Wrong number of arguments.")

    def test_no_tasks(self):
        exc = self.assertRaises(
            UsageError,
            self.mk_opts, [])
        self.assertEqual(str(exc), "Please specify a task.")

    def test_one_task(self):
        opts = self.mk_opts(["-t", "count"])
        self.assertEqual(
            [t.name for t in opts["tasks"]],
            ["count"]
        )

    def test_multiple_tasks(self):
        opts = self.mk_opts(["-t", "list", "-t", "count"])
        self.assertEqual(
            [t.name for t in opts["tasks"]],
            ["list", "count"]
        )

    def test_help(self):
        opts = Options()
        lines = opts.getUsage().splitlines()
        self.assertEqual(lines[-6:], [
            "Available tasks:",
            "      --count    A task that counts the number of keys.",
            "      --expire   A task that sets an expiry time on each key.",
            "      --list     A task that prints out each key.",
            "      --persist  A task that persists each key.",
            "      --skip     A task that skips keys that match a regular"
            " expression.",
        ])


class TestTaskRunner(VumiTestCase):
    def make_runner(self, tasks, redis=None, pattern="*"):
        if redis is None:
            redis = self.mk_redis_config()
        args = tasks + [redis, pattern]
        options = Options()
        options.parseOptions(args)
        runner = TaskRunner(options)
        runner.redis._purge_all()   # Make sure we start fresh.
        runner.stdout = StringIO.StringIO()
        return runner

    def output(self, runner):
        return runner.stdout.getvalue().splitlines()

    def mk_file(self, data):
        name = self.mktemp()
        with open(name, "wb") as data_file:
            data_file.write(data)
        return name

    def mk_redis_config(self):
        config = {
            'redis_manager': {
                'FAKE_REDIS': True,
            },
        }
        return self.mk_file(yaml.safe_dump(config))

    def test_single_task(self):
        runner = self.make_runner([
            "-t", "count",
        ])
        runner.run()
        self.assertEqual(self.output(runner), [
            'Found 0 matching keys.',
        ])

    def test_multiple_task(self):
        runner = self.make_runner([
            "-t", "expire:seconds=10",
            "-t", "count",
        ])
        runner.redis.set("key1", "k1")
        runner.redis.set("key2", "k2")
        runner.run()
        self.assertEqual(self.output(runner), [
            'Found 2 matching keys.',
        ])
        self.assertTrue(0 < runner.redis.ttl("key1") <= 10)
        self.assertTrue(0 < runner.redis.ttl("key2") <= 10)

    def test_match(self):
        runner = self.make_runner([
            "-t", "expire:seconds=10",
            "-t", "count",
        ], pattern="coffee:*")
        runner.redis.set("coffee:key1", "k1")
        runner.redis.set("tea:key2", "k2")
        runner.run()
        self.assertEqual(self.output(runner), [
            'Found 1 matching keys.',
        ])
        self.assertTrue(0 < runner.redis.ttl("coffee:key1") <= 10)
        self.assertEqual(runner.redis.ttl("tea:key2"), None)

    def test_key_skipping(self):
        runner = self.make_runner([
            "-t", "skip:pattern=key1",
            "-t", "list",
        ])
        runner.redis.set("key1", "k1")
        runner.redis.set("key2", "k2")
        runner.run()
        self.assertEqual(self.output(runner), [
            'key2',
        ])


class TestScanKeys(VumiTestCase):
    def setUp(self):
        self.persistence_helper = self.add_helper(
            PersistenceHelper(is_sync=True))
        self.redis = self.persistence_helper.get_redis_manager()
        self.redis._purge_all()  # Make sure we start fresh.

    def test_no_keys(self):
        keys = list(scan_keys(self.redis, "*"))
        self.assertEqual(keys, [])

    def test_single_scan_loop(self):
        expected_keys = ["key%d" % i for i in range(5)]
        for key in expected_keys:
            self.redis.set(key, "foo")
        keys = sorted(scan_keys(self.redis, "*"))
        self.assertEqual(keys, expected_keys)

    def test_multiple_scan_loops(self):
        expected_keys = ["key%02d" % i for i in range(100)]
        for key in expected_keys:
            self.redis.set(key, "foo")
        keys = sorted(scan_keys(self.redis, "*"))
        self.assertEqual(keys, expected_keys)

    def test_match(self):
        self.redis.set("coffee:latte", "yes")
        self.redis.set("tea:rooibos", "yes")
        keys = list(scan_keys(self.redis, "coffee:*"))
        self.assertEqual(keys, ["coffee:latte"])



"""Tests for vumi.middleware.session_length."""

from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.internet.task import Clock

from vumi.message import TransportUserMessage
from vumi.middleware.session_length import (
    SessionLengthMiddleware, SessionLengthMiddlewareError)
from vumi.middleware.tagger import TaggingMiddleware
from vumi.tests.utils import LogCatcher
from vumi.tests.helpers import VumiTestCase, PersistenceHelper

SESSION_NEW, SESSION_CLOSE, SESSION_NONE = (
    TransportUserMessage.SESSION_NEW,
    TransportUserMessage.SESSION_CLOSE,
    TransportUserMessage.SESSION_NONE)


class TestSessionLengthMiddleware(VumiTestCase):
    def setUp(self):
        self.persistence_helper = self.add_helper(PersistenceHelper())
        self.clock = Clock()

    @inlineCallbacks
    def mk_middleware(self, **kw):
        dummy_worker = object()
        config = self.persistence_helper.mk_config(kw)
        mw = SessionLengthMiddleware("session_length", config, dummy_worker)
        yield mw.setup_middleware()
        self.patch(mw, 'clock', self.clock)
        self.redis = mw.redis
        self.add_cleanup(mw.teardown_middleware)
        returnValue(mw)

    def mk_msg(self, to_addr, from_addr, session_event=SESSION_NEW,
               session_start=None, session_end=None, tag=None,
               transport_name='dummy_transport'):
        msg = TransportUserMessage(
            to_addr=to_addr, from_addr=from_addr,
            transport_name=transport_name,
            transport_type="dummy_transport_type",
            session_event=session_event)

        if tag is not None:
            TaggingMiddleware.add_tag_to_msg(msg, tag)

        if session_start is not None:
            self._set_metadata(msg, 'session_start', session_start)

        if session_end is not None:
            self._set_metadata(msg, 'session_end', session_end)

        return msg

    def _set_metadata(self, msg, name, value, metadata_field_name='session'):
        metadata = msg['helper_metadata'].setdefault(metadata_field_name, {})
        metadata[name] = value

    def assert_middleware_error(self, msg):
        [err] = self.flushLoggedErrors(SessionLengthMiddlewareError)
        self.assertEqual(str(err.value), msg)

    @inlineCallbacks
    def test_incoming_message_session_start(self):
        mw = yield self.mk_middleware()
        msg_start = self.mk_msg('+12345', '+54321')

        msg = yield mw.handle_inbound(msg_start, "dummy_connector")
        value = yield self.redis.get('dummy_transport:+54321:session_created')
        self.assertEqual(value, '0.0')

        self.assertEqual(
            msg['helper_metadata']['session']['session_start'], 0.0)

    @inlineCallbacks
    def test_incoming_message_session_end(self):
        mw = yield self.mk_middleware()
        msg_end = self.mk_msg('+12345', '+54321', session_event=SESSION_CLOSE)

        msg = yield mw.handle_inbound(msg_end, "dummy_connector")
        self.assertEqual(
            msg['helper_metadata']['session']['session_end'], 0.0)

    @inlineCallbacks
    def test_incoming_message_session_start_end(self):
        mw = yield self.mk_middleware()
        msg_start = self.mk_msg('+12345', '+54321')
        msg_end = self.mk_msg('+12345', '+54321', session_event=SESSION_CLOSE)

        msg = yield mw.handle_inbound(msg_start, "dummy_connector")
        value = yield self.redis.get('dummy_transport:+54321:session_created')
        self.assertEqual(value, '0.0')

        self.assertEqual(
                msg['helper_metadata']['session']['session_start'], 0.0)

        self.clock.advance(1.0)
        msg = yield mw.handle_inbound(msg_end, "dummy_connector")
        value = yield self.redis.get('dummy_transport:+54321:session_created')
        self.assertTrue(value is None)
        self.assertEqual(
                msg['helper_metadata']['session']['session_start'], 0.0)
        self.assertEqual(
            msg['helper_metadata']['session']['session_end'], 1.0)

    @inlineCallbacks
    def test_incoming_message_session_start_no_overwrite(self):
        mw = yield self.mk_middleware()

        msg = self.mk_msg(
            '+12345',
            '+54321',
            session_start=23,
            session_event=SESSION_NEW)

        yield self.redis.set('dummy_transport:+12345:session_created', '23.0')

        msg = yield mw.handle_inbound(msg, "dummy_connector")
        value = yield self.redis.get('dummy_transport:+12345:session_created')
        self.assertEqual(value, '23.0')

        self.assertEqual(
            msg['helper_metadata']['session']['session_start'], 23.0)

    @inlineCallbacks
    def test_incoming_message_session_end_no_overwrite(self):
        mw = yield self.mk_middleware()

        msg = self.mk_msg(
            '+12345',
            '+54321',
            session_start=23.0,
            session_end=32.0,
            session_event=SESSION_CLOSE)

        msg = yield mw.handle_inbound(msg, "dummy_connector")

        self.assertEqual(
            msg['helper_metadata']['session']['session_start'], 23.0)

        self.assertEqual(
            msg['helper_metadata']['session']['session_end'], 32.0)

    @inlineCallbacks
    def test_incoming_message_session_none_no_overwrite(self):
        mw = yield self.mk_middleware()

        msg = self.mk_msg(
            '+12345',
            '+54321',
            session_start=23.0,
            session_event=SESSION_NONE)

        msg = yield mw.handle_inbound(msg, "dummy_connector")

        self.assertEqual(
            msg['helper_metadata']['session']['session_start'], 23.0)

    @inlineCallbacks
    def test_incoming_message_session_start_transport_namespace_type(self):
        mw = yield self.mk_middleware(namespace_type='transport_name')

        msg = self.mk_msg('+12345', '+54321', transport_name='foo')
        msg = yield mw.handle_inbound(msg, "dummy_connector")

        value = yield self.redis.get('foo:+54321:session_created')
        self.assertEqual(value, '0.0')

        self.assertEqual(
            msg['helper_metadata']['session']['session_start'], 0.0)

    @inlineCallbacks
    def test_incoming_message_session_end_transport_namespace_type(self):
        mw = yield self.mk_middleware(namespace_type='transport_name')
        yield self.redis.set('foo:+54321:session_created', '23.0')

        msg = self.mk_msg(
            '+12345',
            '+54321',
            session_event=SESSION_CLOSE,
            transport_name='foo')

        msg = yield mw.handle_inbound(msg, "dummy_connector")

        value = yield self.redis.get('foo:+54321:session_created')
        self.assertEqual(value, None)

        self.assertEqual(
            msg['helper_metadata']['session']['session_start'], 23.0)

        self.assertEqual(
            msg['helper_metadata']['session']['session_end'], 0.0)

    @inlineCallbacks
    def test_incoming_message_session_none_transport_namespace_type(self):
        mw = yield self.mk_middleware(namespace_type='transport_name')
        yield self.redis.set('foo:+54321:session_created', '23.0')

        msg = self.mk_msg(
            '+12345',
            '+54321',
            session_event=SESSION_CLOSE,
            transport_name='foo')

        msg = yield mw.handle_inbound(msg, "dummy_connector")

        self.assertEqual(
            msg['helper_metadata']['session']['session_start'], 23.0)

    @inlineCallbacks
    def test_incoming_message_session_start_no_transport_name(self):
        mw = yield self.mk_middleware(namespace_type='transport_name')
        msg = self.mk_msg('+12345', '+54321', transport_name=None)

        result_msg = yield mw.handle_inbound(msg, "dummy_connector")
        self.assert_middleware_error(
            "Session length redis namespace cannot be None, skipping message")

        self.assertEqual(msg, result_msg)
        self.assertTrue('session' not in msg['helper_metadata'])

    @inlineCallbacks
    def test_incoming_message_session_end_no_transport_name(self):
        mw = yield self.mk_middleware(namespace_type='transport_name')

        msg = self.mk_msg(
            '+12345',
            '+54321',
            session_event=SESSION_CLOSE,
            transport_name=None)

        result_msg = yield mw.handle_inbound(msg, "dummy_connector")
        self.assert_middleware_error(
            "Session length redis namespace cannot be None, skipping message")

        self.assertEqual(msg, result_msg)
        self.assertTrue('session' not in msg['helper_metadata'])

    @inlineCallbacks
    def test_incoming_message_session_none_no_transport_name(self):
        mw = yield self.mk_middleware(namespace_type='transport_name')

        msg = self.mk_msg(
            '+12345',
            '+54321',
            session_event=SESSION_NONE,
            transport_name=None)

        result_msg = yield mw.handle_inbound(msg, "dummy_connector")
        self.assert_middleware_error(
            "Session length redis namespace cannot be None, skipping message")

        self.assertEqual(msg, result_msg)
        self.assertTrue('session' not in msg['helper_metadata'])

    @inlineCallbacks
    def test_incoming_message_session_start_tag_namespace_type(self):
        mw = yield self.mk_middleware(namespace_type='tag')

        msg = self.mk_msg('+12345', '+54321', tag=('pool1', 'tag1'))
        msg = yield mw.handle_inbound(msg, "dummy_connector")

        value = yield self.redis.get('pool1:tag1:+54321:session_created')
        self.assertEqual(value, '0.0')

        self.assertEqual(
            msg['helper_metadata']['session']['session_start'], 0.0)

    @inlineCallbacks
    def test_incoming_message_session_end_tag_namespace_type(self):
        mw = yield self.mk_middleware(namespace_type='tag')
        yield self.redis.set('pool1:tag1:+54321:session_created', '23.0')

        msg = self.mk_msg(
            '+12345',
            '+54321',
            session_event=SESSION_CLOSE,
            tag=('pool1', 'tag1'))

        msg = yield mw.handle_inbound(msg, "dummy_connector")

        value = yield self.redis.get('pool1:tag1:+54321:session_created')
        self.assertEqual(value, None)

        self.assertEqual(
            msg['helper_metadata']['session']['session_start'], 23.0)

        self.assertEqual(
            msg['helper_metadata']['session']['session_end'], 0.0)

    @inlineCallbacks
    def test_incoming_message_session_none_tag_namespace_type(self):
        mw = yield self.mk_middleware(namespace_type='tag')
        yield self.redis.set('pool1:tag1:+54321:session_created', '23.0')

        msg = self.mk_msg(
            '+12345',
            '+54321',
            session_event=SESSION_CLOSE,
            tag=('pool1', 'tag1'))

        msg = yield mw.handle_inbound(msg, "dummy_connector")

        self.assertEqual(
            msg['helper_metadata']['session']['session_start'], 23.0)

    @inlineCallbacks
    def test_incoming_message_session_start_no_tag(self):
        mw = yield self.mk_middleware(namespace_type='tag')

        # create message with no tag
        msg = self.mk_msg('+12345', '+54321')

        result_msg = yield mw.handle_inbound(msg, "dummy_connector")
        self.assert_middleware_error(
            "Session length redis namespace cannot be None, skipping message")

        self.assertEqual(msg, result_msg)
        self.assertTrue('session' not in msg['helper_metadata'])

    @inlineCallbacks
    def test_incoming_message_session_end_tag_no_tag(self):
        mw = yield self.mk_middleware(namespace_type='tag')

        # create message with no tag
        msg = self.mk_msg('+12345', '+54321', session_event=SESSION_CLOSE)

        result_msg = yield mw.handle_inbound(msg, "dummy_connector")
        self.assert_middleware_error(
            "Session length redis namespace cannot be None, skipping message")

        self.assertEqual(msg, result_msg)
        self.assertTrue('session' not in msg['helper_metadata'])

    @inlineCallbacks
    def test_incoming_message_session_none_tag_no_tag(self):
        mw = yield self.mk_middleware(namespace_type='tag')
        yield self.redis.set('pool1:tag1:+54321:session_created', '23.0')

        # create message with no tag
        msg = self.mk_msg('+12345', '+54321', session_event=SESSION_NONE)

        result_msg = yield mw.handle_inbound(msg, "dummy_connector")
        self.assert_middleware_error(
            "Session length redis namespace cannot be None, skipping message")

        self.assertEqual(msg, result_msg)
        self.assertTrue('session' not in msg['helper_metadata'])

    @inlineCallbacks
    def test_outgoing_message_session_start(self):
        mw = yield self.mk_middleware()
        msg_start = self.mk_msg('+12345', '+54321')

        msg = yield mw.handle_outbound(msg_start, "dummy_connector")
        value = yield self.redis.get('dummy_transport:+12345:session_created')
        self.assertEqual(value, '0.0')
        self.assertEqual(
                msg['helper_metadata']['session']['session_start'], 0.0)

    @inlineCallbacks
    def test_outgoing_message_session_end(self):
        mw = yield self.mk_middleware()
        msg_end = self.mk_msg('+12345', '+54321', session_event=SESSION_CLOSE)

        msg = yield mw.handle_outbound(msg_end, "dummy_connector")
        self.assertEqual(
                msg['helper_metadata']['session']['session_end'], 0.0)

    @inlineCallbacks
    def test_outgoing_message_session_start_end(self):
        mw = yield self.mk_middleware()
        msg_start = self.mk_msg('+12345', '+54321')
        msg_end = self.mk_msg('+12345', '+54321', session_event=SESSION_CLOSE)

        msg = yield mw.handle_outbound(msg_start, "dummy_connector")
        value = yield self.redis.get('dummy_transport:+12345:session_created')
        self.assertEqual(value, '0.0')
        self.assertEqual(
                msg['helper_metadata']['session']['session_start'], 0.0)

        self.clock.advance(1.0)
        msg = yield mw.handle_outbound(msg_end, "dummy_connector")
        value = yield self.redis.get('dummy_transport:+54321:session_created')
        self.assertTrue(value is None)
        self.assertEqual(
                msg['helper_metadata']['session']['session_start'], 0.0)
        self.assertEqual(
            msg['helper_metadata']['session']['session_end'], 1.0)

    @inlineCallbacks
    def test_outgoing_message_session_start_no_overwrite(self):
        mw = yield self.mk_middleware()

        msg = self.mk_msg(
            '+12345',
            '+54321',
            session_start=23,
            session_event=SESSION_NEW)

        yield self.redis.set('dummy_transport:+12345:session_created', '23')

        msg = yield mw.handle_outbound(msg, "dummy_connector")
        value = yield self.redis.get('dummy_transport:+12345:session_created')
        self.assertEqual(value, '23')

        self.assertEqual(
            msg['helper_metadata']['session']['session_start'], 23)

    @inlineCallbacks
    def test_outgoing_message_session_end_no_overwrite(self):
        mw = yield self.mk_middleware()

        msg = self.mk_msg(
            '+12345',
            '+54321',
            session_start=23,
            session_end=32,
            session_event=SESSION_CLOSE)

        msg = yield mw.handle_outbound(msg, "dummy_connector")

        self.assertEqual(
            msg['helper_metadata']['session']['session_start'], 23)

        self.assertEqual(
            msg['helper_metadata']['session']['session_end'], 32)

    @inlineCallbacks
    def test_outgoing_message_session_none_no_overwrite(self):
        mw = yield self.mk_middleware()

        msg = self.mk_msg(
            '+12345',
            '+54321',
            session_start=23,
            session_event=SESSION_NONE)

        msg = yield mw.handle_outbound(msg, "dummy_connector")

        self.assertEqual(
            msg['helper_metadata']['session']['session_start'], 23)

    @inlineCallbacks
    def test_outgoing_message_session_start_no_transport_name(self):
        mw = yield self.mk_middleware(namespace_type='transport_name')

        msg = self.mk_msg('+12345', '+54321', transport_name=None)

        result_msg = yield mw.handle_outbound(msg, "dummy_connector")
        self.assert_middleware_error(
            "Session length redis namespace cannot be None, skipping message")

        self.assertEqual(msg, result_msg)
        self.assertTrue('session' not in msg['helper_metadata'])

    @inlineCallbacks
    def test_outgoing_message_session_end_no_transport_name(self):
        mw = yield self.mk_middleware(namespace_type='transport_name')

        msg = self.mk_msg(
            '+12345',
            '+54321',
            session_event=SESSION_CLOSE,
            transport_name=None)

        result_msg = yield mw.handle_outbound(msg, "dummy_connector")
        self.assert_middleware_error(
            "Session length redis namespace cannot be None, skipping message")

        self.assertEqual(msg, result_msg)
        self.assertTrue('session' not in msg['helper_metadata'])

    @inlineCallbacks
    def test_outgoing_message_session_none_no_transport_name(self):
        mw = yield self.mk_middleware(namespace_type='transport_name')

        msg = self.mk_msg(
            '+12345',
            '+54321',
            session_event=SESSION_NONE,
            transport_name=None)

        result_msg = yield mw.handle_outbound(msg, "dummy_connector")
        self.assert_middleware_error(
            "Session length redis namespace cannot be None, skipping message")

        self.assertEqual(msg, result_msg)
        self.assertTrue('session' not in msg['helper_metadata'])

    @inlineCallbacks
    def test_outgoing_message_session_start_tag_namespace_type(self):
        mw = yield self.mk_middleware(namespace_type='tag')

        msg = self.mk_msg('+12345', '+54321', tag=('pool1', 'tag1'))
        msg = yield mw.handle_outbound(msg, "dummy_connector")

        value = yield self.redis.get('pool1:tag1:+12345:session_created')
        self.assertEqual(value, '0.0')

        self.assertEqual(
            msg['helper_metadata']['session']['session_start'], 0.0)

    @inlineCallbacks
    def test_outgoing_message_session_end_tag_namespace_type(self):
        mw = yield self.mk_middleware(namespace_type='tag')
        yield self.redis.set('pool1:tag1:+12345:session_created', '23.0')

        msg = self.mk_msg(
            '+12345',
            '+54321',
            session_event=SESSION_CLOSE,
            tag=('pool1', 'tag1'))

        msg = yield mw.handle_outbound(msg, "dummy_connector")

        value = yield self.redis.get('tag1:+12345:session_created')
        self.assertEqual(value, None)

        self.assertEqual(
            msg['helper_metadata']['session']['session_start'], 23.0)

        self.assertEqual(
            msg['helper_metadata']['session']['session_end'], 0.0)

    @inlineCallbacks
    def test_outgoing_message_session_none_tag_namespace_type(self):
        mw = yield self.mk_middleware(namespace_type='tag')
        yield self.redis.set('pool1:tag1:+12345:session_created', '23.0')

        msg = self.mk_msg(
            '+12345',
            '+54321',
            session_event=SESSION_NONE,
            tag=('pool1', 'tag1'))

        msg = yield mw.handle_outbound(msg, "dummy_connector")

        self.assertEqual(
            msg['helper_metadata']['session']['session_start'], 23.0)

    @inlineCallbacks
    def test_outgoing_message_session_start_no_tag(self):
        mw = yield self.mk_middleware(namespace_type='tag')

        # create message with no tag
        msg = self.mk_msg('+12345', '+54321')

        result_msg = yield mw.handle_outbound(msg, "dummy_connector")
        self.assert_middleware_error(
            "Session length redis namespace cannot be None, skipping message")

        self.assertEqual(msg, result_msg)
        self.assertTrue('session' not in msg['helper_metadata'])

    @inlineCallbacks
    def test_outgoing_message_session_end_tag_no_tag(self):
        mw = yield self.mk_middleware(namespace_type='tag')

        # create message with no tag
        msg = self.mk_msg('+12345', '+54321', session_event=SESSION_CLOSE)

        result_msg = yield mw.handle_outbound(msg, "dummy_connector")
        self.assert_middleware_error(
            "Session length redis namespace cannot be None, skipping message")

        self.assertEqual(msg, result_msg)
        self.assertTrue('session' not in msg['helper_metadata'])

    @inlineCallbacks
    def test_outgoing_message_session_none_tag_no_tag(self):
        mw = yield self.mk_middleware(namespace_type='tag')
        yield self.redis.set('pool1:tag1:+54321:session_created', '23.0')

        # create message with no tag
        msg = self.mk_msg('+12345', '+54321', session_event=SESSION_NONE)

        result_msg = yield mw.handle_outbound(msg, "dummy_connector")
        self.assert_middleware_error(
            "Session length redis namespace cannot be None, skipping message")

        self.assertEqual(msg, result_msg)
        self.assertTrue('session' not in msg['helper_metadata'])

    @inlineCallbacks
    def test_redis_key_timeout(self):
        mw = yield self.mk_middleware()
        msg_start = self.mk_msg('+12345', '+54321')

        yield mw.handle_inbound(msg_start, "dummy_connector")
        ttl = yield self.redis.ttl('dummy_transport::+54321:session_created')
        self.assertTrue(ttl <= 120)

    @inlineCallbacks
    def test_redis_key_custom_timeout(self):
        mw = yield self.mk_middleware(timeout=20)
        msg_start = self.mk_msg('+12345', '+54321')

        yield mw.handle_inbound(msg_start, "dummy_connector")
        ttl = yield self.redis.ttl('dummy_transport:+54321:session_created')
        self.assertTrue(ttl <= 20)

    @inlineCallbacks
    def test_custom_message_field_name(self):
        mw = yield self.mk_middleware(field_name='foobar')
        msg_start = self.mk_msg('+12345', '+54321')

        msg = yield mw.handle_inbound(msg_start, "dummy_connector")
        self.assertEqual(
            msg['helper_metadata']['foobar']['session_start'], 0.0)

    @inlineCallbacks
    def test_incoming_message_session_no_from_addr(self):
        mw = yield self.mk_middleware()
        msg_start = self.mk_msg('+12345', None)
        msg = yield mw.handle_inbound(msg_start, "dummy_connector")
        self.assertEqual(msg, msg_start)
        self.assert_middleware_error(
            "Session length key address cannot be None, skipping message")

    @inlineCallbacks
    def test_outgoing_message_session_no_to_addr(self):
        mw = yield self.mk_middleware()
        msg_start = self.mk_msg(None, '+54321')
        msg = yield mw.handle_outbound(msg_start, "dummy_connector")
        self.assertEqual(msg, msg_start)
        self.assert_middleware_error(
            "Session length key address cannot be None, skipping message")



# -*- test-case-name: vumi.components.tests.test_session -*-

"""Session management utilities."""

import time

from twisted.internet.defer import inlineCallbacks, returnValue

from vumi import log


class SessionManager(object):
    """A manager for sessions.

    :param TxRedisManager redis:
        Redis manager object.
    :param int max_session_length:
        Time before a session expires. Default is None (never expire).
    :param float gc_period:
        Deprecated and ignored.
    """

    def __init__(self, redis, max_session_length=None, gc_period=None):
        self.max_session_length = max_session_length
        self.redis = redis
        if gc_period is not None:
            log.warning("SessionManager 'gc_period' parameter is deprecated.")

    @inlineCallbacks
    def stop(self, stop_redis=True):
        if stop_redis:
            yield self.redis._close()

    @classmethod
    def from_redis_config(cls, config, key_prefix=None,
                          max_session_length=None, gc_period=None):
        """Create a `SessionManager` instance using `TxRedisManager`.
        """
        from vumi.persist.txredis_manager import TxRedisManager
        d = TxRedisManager.from_config(config)
        if key_prefix is not None:
            d.addCallback(lambda m: m.sub_manager(key_prefix))
        return d.addCallback(lambda m: cls(m, max_session_length, gc_period))

    @inlineCallbacks
    def active_sessions(self):
        """Return a list of active user_ids and associated sessions.

        Queries redis for keys starting with the session key prefix. This is
        O(n) over the total number of keys in redis, but this is still pretty
        quick even for millions of keys. Try not to hit this too often, though.
        """
        keys = yield self.redis.keys('session:*')
        sessions = []
        for user_id in [key.split(':', 1)[1] for key in keys]:
            sessions.append((user_id, (yield self.load_session(user_id))))

        returnValue(sessions)

    def load_session(self, user_id):
        """
        Load session data from Redis
        """
        ukey = "%s:%s" % ('session', user_id)
        return self.redis.hgetall(ukey)

    def schedule_session_expiry(self, user_id, timeout):
        """
        Schedule a session to timeout

        Parameters
        ----------
        user_id : str
            The user's id.
        timeout : int
            The number of seconds after which this session should expire
        """
        ukey = "%s:%s" % ('session', user_id)
        return self.redis.expire(ukey, timeout)

    @inlineCallbacks
    def create_session(self, user_id, **kwargs):
        """
        Create a new session using the given user_id
        """
        yield self.clear_session(user_id)
        defaults = {
            'created_at': time.time()
        }
        defaults.update(kwargs)
        yield self.save_session(user_id, defaults)
        if self.max_session_length:
            yield self.schedule_session_expiry(user_id,
                                               int(self.max_session_length))
        returnValue((yield self.load_session(user_id)))

    def clear_session(self, user_id):
        ukey = "%s:%s" % ('session', user_id)
        return self.redis.delete(ukey)

    @inlineCallbacks
    def save_session(self, user_id, session):
        """
        Save a session

        Parameters
        ----------
        user_id : str
            The user's id.
        session : dict
            The session info, nested dictionaries are not supported. Any
            values that are dictionaries are converted to strings by Redis.

        """
        ukey = "%s:%s" % ('session', user_id)
        for s_key, s_value in session.items():
            yield self.redis.hset(ukey, s_key, s_value)
        returnValue(session)



# -*- test-case-name: vumi.transports.parlayx.tests.test_client -*-
import hashlib
import uuid
from collections import namedtuple
from datetime import datetime

from vumi.transports.parlayx.soaputil import perform_soap_request, SoapFault
from vumi.transports.parlayx.xmlutil import (
    gettext, gettextall, Namespace, LocalNamespace as L)


PARLAYX_COMMON_NS = Namespace(
    'http://www.csapi.org/schema/parlayx/common/v2_1', 'parlayx_common')
SEND_NS = Namespace(
    'http://www.csapi.org/schema/parlayx/sms/send/v2_2/local', 'send')
NOTIFICATION_MANAGER_NS = Namespace(
    'http://www.csapi.org/schema/parlayx/sms/notification_manager/v2_3/local',
    'nm')
PARLAYX_HEAD_NS = Namespace(
    'http://www.huawei.com.cn/schema/common/v2_1', 'parlayx_head')


def format_address(msisdn):
    """
    Format a normalized MSISDN as a URI that ParlayX will accept.
    """
    if not msisdn.startswith('+'):
        raise ValueError('Only international format addresses are supported')
    return 'tel:' + msisdn[1:]


def format_timestamp(when):
    """
    Format a `datetime` instance timestamp according to ParlayX
    requirements.
    """
    return when.strftime('%Y%m%d%H%M%S')


def make_password(service_provider_id, service_provider_password,
                   timestamp):
    """
    Build a time-sensitive password for a request.
    """
    return hashlib.md5(
        service_provider_id +
        service_provider_password +
        timestamp).hexdigest()


class _ParlayXFaultDetail(namedtuple('_ParlayXFaultDetail',
                                   ['message_id', 'text', 'variables'])):
    """
    Generic ParlayX SOAP fault detail.
    """
    tag = None

    @classmethod
    def from_element(cls, element):
        if element.tag != cls.tag.text:
            return None
        return cls(
            message_id=gettext(element, 'messageId'),
            text=gettext(element, 'text'),
            variables=list(gettextall(element, 'variables')))


class ServiceExceptionDetail(_ParlayXFaultDetail):
    """
    ParlayX service exception detail.
    """
    tag = PARLAYX_COMMON_NS.ServiceExceptionDetail


class ServiceException(SoapFault):
    """
    ParlayX service exception.
    """
    detail_type = ServiceExceptionDetail


class PolicyExceptionDetail(_ParlayXFaultDetail):
    """
    ParlayX policy exception detail.
    """
    tag = PARLAYX_COMMON_NS.PolicyExceptionDetail


class PolicyException(SoapFault):
    """
    ParlayX policy exception.
    """
    detail_type = PolicyExceptionDetail


class ParlayXClient(object):
    """
    ParlayX SOAP client.

    :ivar _service_correlator:
        A unique identifier for this service, used when registering and
        deregistering for SMS notifications.
    """
    def __init__(self, service_provider_service_id, service_provider_id,
                 service_provider_password, short_code, endpoint, send_uri,
                 notification_uri, perform_soap_request=perform_soap_request):
        """
        :param service_provider_service_id:
            Provisioned service provider service identifier.
        :param service_provider_id:
            Provisioned service provider identifier/username.
        :param service_provider_password:
            Provisioned service provider password.
        :param short_code:
            SMS shortcode or service activation number.
        :param endpoint:
            URI to which the remote ParlayX service will deliver notification
            messages.
        :param send_uri:
            URI for the ParlayX ``SendSmsService`` SOAP endpoint.
        :param notification_uri:
            URI for the ParlayX ``SmsNotificationService`` SOAP endpoint.
        """
        self.service_provider_service_id = service_provider_service_id
        self.service_provider_id = service_provider_id
        self.service_provider_password = service_provider_password
        self.short_code = short_code
        self.endpoint = endpoint
        self.send_uri = send_uri
        self.notification_uri = notification_uri
        self.perform_soap_request = perform_soap_request
        self._service_correlator = uuid.uuid4().hex

    def _now(self):
        """
        The current date and time.
        """
        return datetime.now()

    def _make_header(self, service_subscription_address=None, linkid=None):
        """
        Create a ``RequestSOAPHeader`` element.

        :param service_subscription_address:
            Service subscription address for the ``OA`` header field, this
            field is omitted if its value is ``None``.
        """
        NS = PARLAYX_HEAD_NS
        other = []
        timestamp = format_timestamp(self._now())
        if service_subscription_address is not None:
            other.append(NS.OA(format_address(service_subscription_address)))
        if linkid is not None:
            other.append(NS.linkid(linkid))
        return NS.RequestSOAPHeader(
            NS.spId(self.service_provider_id),
            NS.spPassword(
                make_password(
                    self.service_provider_id,
                    self.service_provider_password,
                    timestamp)),
            NS.serviceId(self.service_provider_service_id),
            NS.timeStamp(timestamp),
            *other)

    def start_sms_notification(self):
        """
        Register a notification delivery endpoint with the remote ParlayX
        service.
        """
        body = NOTIFICATION_MANAGER_NS.startSmsNotification(
            NOTIFICATION_MANAGER_NS.reference(
                L.endpoint(self.endpoint),
                L.interfaceName('notifySmsReception'),
                L.correlator(self._service_correlator)),
            NOTIFICATION_MANAGER_NS.smsServiceActivationNumber(
                self.short_code))
        header = self._make_header()
        return self.perform_soap_request(
            uri=self.notification_uri,
            action='',
            body=body,
            header=header,
            expected_faults=[ServiceException])

    def stop_sms_notification(self):
        """
        Deregister notification delivery with the remote ParlayX service.
        """
        body = NOTIFICATION_MANAGER_NS.stopSmsNotification(
            L.correlator(self._service_correlator))
        header = self._make_header()
        return self.perform_soap_request(
            uri=self.notification_uri,
            action='',
            body=body,
            header=header,
            expected_faults=[ServiceException])

    def send_sms(self, to_addr, content, message_id, linkid=None):
        """
        Send an SMS.
        """
        def _extractRequestIdentifier((body, header)):
            return gettext(body, './/' + str(SEND_NS.result), default='')

        body = SEND_NS.sendSms(
            SEND_NS.addresses(format_address(to_addr)),
            SEND_NS.message(content),
            SEND_NS.receiptRequest(
                L.endpoint(self.endpoint),
                L.interfaceName(u'SmsNotification'),
                L.correlator(message_id)))
        header = self._make_header(
            service_subscription_address=to_addr,
            linkid=linkid)
        d = self.perform_soap_request(
            uri=self.send_uri,
            action='',
            body=body,
            header=header,
            expected_faults=[PolicyException, ServiceException])
        d.addCallback(_extractRequestIdentifier)
        return d


__all__ = [
    'format_address', 'ServiceExceptionDetail', 'ServiceException',
    'PolicyExceptionDetail', 'PolicyException', 'ParlayXClient']



# -*- test-case-name: vumi.transports.smpp.deprecated.clientserver.test.test_server -*-

import uuid
from datetime import datetime

from twisted.python import log
from twisted.internet import reactor
from twisted.internet.protocol import Protocol, ServerFactory

from smpp.pdu_builder import (BindTransceiverResp,
                                BindTransmitterResp,
                                BindReceiverResp,
                                EnquireLinkResp,
                                SubmitSMResp,
                                DeliverSM)
from smpp.pdu_inspector import binascii, unpack_pdu


class SmscServer(Protocol):

    def __init__(self, delivery_report_string=None):
        log.msg('__init__', 'SmscServer')
        self.delivery_report_string = delivery_report_string
        if self.delivery_report_string is None:
            self.delivery_report_string = 'id:%' \
                    's sub:001 dlvrd:001 submit date:%' \
                    's done date:%' \
                    's stat:DELIVRD err:000 text:'
        self.datastream = ''

    def pop_data(self):
        data = None
        if(len(self.datastream) >= 16):
            command_length = int(binascii.b2a_hex(self.datastream[0:4]), 16)
            if(len(self.datastream) >= command_length):
                data = self.datastream[0:command_length]
                self.datastream = self.datastream[command_length:]
        return data

    def handle_data(self, data):
        pdu = unpack_pdu(data)
        log.msg('INCOMING <<<< %r' % (pdu,))
        if pdu['header']['command_id'] == 'bind_transceiver':
            self.handle_bind_transceiver(pdu)
        if pdu['header']['command_id'] == 'bind_transmitter':
            self.handle_bind_transmitter(pdu)
        if pdu['header']['command_id'] == 'bind_receiver':
            self.handle_bind_receiver(pdu)
        if pdu['header']['command_id'] == 'submit_sm':
            self.handle_submit_sm(pdu)
        if pdu['header']['command_id'] == 'enquire_link':
            self.handle_enquire_link(pdu)

    def handle_bind_transceiver(self, pdu):
        if pdu['header']['command_status'] == 'ESME_ROK':
            sequence_number = pdu['header']['sequence_number']
            system_id = pdu['body']['mandatory_parameters']['system_id']
            pdu_resp = BindTransceiverResp(sequence_number,
                                           system_id=system_id)
            self.send_pdu(pdu_resp)

    def handle_bind_transmitter(self, pdu):
        if pdu['header']['command_status'] == 'ESME_ROK':
            sequence_number = pdu['header']['sequence_number']
            system_id = pdu['body']['mandatory_parameters']['system_id']
            pdu_resp = BindTransmitterResp(sequence_number,
                                           system_id=system_id)
            self.send_pdu(pdu_resp)

    def handle_bind_receiver(self, pdu):
        if pdu['header']['command_status'] == 'ESME_ROK':
            sequence_number = pdu['header']['sequence_number']
            system_id = pdu['body']['mandatory_parameters']['system_id']
            pdu_resp = BindReceiverResp(sequence_number,
                                        system_id=system_id)
            self.send_pdu(pdu_resp)

    def handle_enquire_link(self, pdu):
        if pdu['header']['command_status'] == 'ESME_ROK':
            sequence_number = pdu['header']['sequence_number']
            pdu_resp = EnquireLinkResp(sequence_number)
            self.send_pdu(pdu_resp)

    def command_status(self, pdu):
        if pdu['body']['mandatory_parameters']['short_message'][:5] == "ESME_":
            return pdu['body']['mandatory_parameters']['short_message'].split(
                    ' ')[0]
        else:
            return 'ESME_ROK'

    def handle_submit_sm(self, pdu):
        if pdu['header']['command_status'] == 'ESME_ROK':
            sequence_number = pdu['header']['sequence_number']
            message_id = str(uuid.uuid4())
            command_status = self.command_status(pdu)
            pdu_resp = SubmitSMResp(
                    sequence_number, message_id, command_status)
            self.send_pdu(pdu_resp)
            reactor.callLater(0, self.delivery_report, message_id)

    def delivery_report(self, message_id):
        sequence_number = 1
        short_message = (self.delivery_report_string % (
                         message_id, datetime.now().strftime("%y%m%d%H%M%S"),
                         datetime.now().strftime("%y%m%d%H%M%S")))
        pdu = DeliverSM(sequence_number, short_message=short_message)
        self.send_pdu(pdu)

    def dataReceived(self, data):
        self.datastream += data
        data = self.pop_data()
        while data is not None:
            self.handle_data(data)
            data = self.pop_data()

    def send_pdu(self, pdu):
        data = pdu.get_bin()
        log.msg('OUTGOING >>>> %r' % (unpack_pdu(data),))
        self.transport.write(data)


class SmscServerFactory(ServerFactory):
    protocol = SmscServer

    def __init__(self, delivery_report_string=None):
        self.delivery_report_string = delivery_report_string

    def buildProtocol(self, addr):
        self.smsc = self.protocol(self.delivery_report_string)
        return self.smsc



# -*- test-case-name: vumi.demos.tests.test_ircbot -*-

"""Demo workers for constructing a simple IRC bot."""

import re
import json

from twisted.internet.defer import inlineCallbacks, returnValue

from vumi import log
from vumi.application import ApplicationWorker
from vumi.persist.txredis_manager import TxRedisManager
from vumi.config import ConfigText, ConfigDict


class MemoConfig(ApplicationWorker.CONFIG_CLASS):
    "Memo worker config."
    worker_name = ConfigText("Name of worker.", required=True, static=True)
    redis_manager = ConfigDict(
        "Redis client configuration.", default={}, static=True)


class MemoWorker(ApplicationWorker):
    """Watches for memos to users and notifies users of memos when users
    appear.
    """

    CONFIG_CLASS = MemoConfig

    MEMO_RE = re.compile(r'^\S+ tell (\S+) (.*)$')

    @inlineCallbacks
    def setup_application(self):
        config = self.get_static_config()
        r_prefix = "ircbot:memos:%s" % (config.worker_name,)
        redis = yield TxRedisManager.from_config(config.redis_manager)
        self.redis = redis.sub_manager(r_prefix)

    def teardown_application(self):
        return self.redis._close()

    def rkey_memo(self, channel, recipient):
        return "%s:%s" % (channel, recipient)

    def store_memo(self, channel, recipient, sender, text):
        memo_key = self.rkey_memo(channel, recipient)
        value = json.dumps([sender, text])
        return self.redis.rpush(memo_key, value)

    @inlineCallbacks
    def retrieve_memos(self, channel, recipient, delete=False):
        memo_key = self.rkey_memo(channel, recipient)
        memos = yield self.redis.lrange(memo_key, 0, -1)
        if delete:
            yield self.redis.delete(memo_key)
        returnValue([json.loads(value) for value in memos])

    @inlineCallbacks
    def consume_user_message(self, msg):
        """Log message from a user."""
        nickname = msg.user()
        irc_metadata = msg['helper_metadata'].get('irc', {})
        channel = irc_metadata.get('irc_channel', 'unknown')
        addressed_to = irc_metadata.get('addressed_to_transport', True)

        if addressed_to:
            yield self.process_potential_memo(channel, nickname, msg)

        memos = yield self.retrieve_memos(channel, nickname, delete=True)
        if memos:
            log.msg("Time to deliver some memos:", memos)
        for memo_sender, memo_text in memos:
            self.reply_to(msg, "%s, %s asked me tell you: %s"
                          % (nickname, memo_sender, memo_text))

    @inlineCallbacks
    def process_potential_memo(self, channel, sender, msg):
        match = self.MEMO_RE.match(msg['content'])
        if match:
            recipient = match.group(1).lower()
            memo_text = match.group(2)
            yield self.store_memo(channel, recipient, sender, memo_text)
            self.reply_to(msg, "%s: Sure thing, boss." % (sender,))



import iso8601
from datetime import datetime
from StringIO import StringIO

from twisted.trial.unittest import TestCase
from twisted.web import http
from twisted.web.test.requesthelper import DummyRequest

from vumi.transports.parlayx.server import (
    NOTIFICATION_NS, PARLAYX_COMMON_NS, normalize_address, SmsMessage,
    DeliveryInformation, DeliveryStatus, SmsNotificationService)
from vumi.transports.parlayx.soaputil import SoapFault, SOAP_ENV, soap_envelope
from vumi.transports.parlayx.xmlutil import (
    ParseError, LocalNamespace as L, tostring, fromstring, element_to_dict)
from vumi.transports.parlayx.tests.utils import (
    create_sms_reception_element, create_sms_delivery_receipt)


class NormalizeAddressTests(TestCase):
    """
    Tests for `vumi.transports.parlayx.server.normalize_address`.
    """
    def test_not_prefixed(self):
        """
        `normalize_address` still normalizes addresses that are not prefixed
        with ``tel:``.
        """
        self.assertEqual(
            '+27117654321', normalize_address('27 11 7654321'))
        self.assertEqual(
            '54321', normalize_address('54321'))

    def test_prefixed(self):
        """
        `normalize_address` strips any ``tel:`` prefix and normalizes the
        address.
        """
        self.assertEqual(
            '+27117654321', normalize_address('tel:27 11 7654321'))
        self.assertEqual(
            '+27117654321', normalize_address('tel:27117654321'))
        self.assertEqual(
            '54321', normalize_address('tel:54321'))


class SmsMessageTests(TestCase):
    """
    Tests for `vumi.transports.parlayx.server.SmsMessage`.
    """
    def test_from_element(self):
        """
        `SmsMessage.from_element` parses a ParlayX ``SmsMessage`` complex type,
        with an ISO8601 timestamp, into an `SmsMessage` instance.
        """
        timestamp = datetime(
            2013, 6, 12, 13, 15, 0, tzinfo=iso8601.iso8601.Utc())
        msg = SmsMessage.from_element(
            NOTIFICATION_NS.message(
                L.message('message'),
                L.senderAddress('tel:27117654321'),
                L.smsServiceActivationNumber('54321'),
                L.dateTime('2013-06-12T13:15:00')))
        self.assertEqual(
            ('message', '+27117654321', '54321', timestamp),
            (msg.message, msg.sender_address, msg.service_activation_number,
             msg.timestamp))

    def test_from_element_missing_timestamp(self):
        """
        `SmsMessage.from_element` parses a ParlayX ``SmsMessage`` complex type,
        without a timestamp, into an `SmsMessage` instance.
        """
        msg = SmsMessage.from_element(
            NOTIFICATION_NS.message(
                L.message('message'),
                L.senderAddress('tel:27117654321'),
                L.smsServiceActivationNumber('54321')))
        self.assertEqual(
            ('message', '+27117654321', '54321', None),
            (msg.message, msg.sender_address, msg.service_activation_number,
             msg.timestamp))


class DeliveryInformationTests(TestCase):
    """
    Tests for `vumi.transports.parlayx.server.DeliveryInformation`.
    """
    def test_from_element(self):
        """
        `DeliveryInformation.from_element` parses a ParlayX
        ``DeliveryInformation`` complex type into a `DeliveryInformation`
        instance. Known ``DeliveryStatus`` enumeration values are parsed into
        `DeliveryStatus` attributes.
        """
        info = DeliveryInformation.from_element(
            NOTIFICATION_NS.deliveryStatus(
                L.address('tel:27117654321'),
                L.deliveryStatus('DeliveredToNetwork')))
        self.assertEqual(
            ('+27117654321', DeliveryStatus.DeliveredToNetwork),
            (info.address, info.delivery_status))

    def test_from_element_unknown_status(self):
        """
        `DeliveryInformation.from_element` raises ``ValueError`` if an unknown
        ``DeliveryStatus`` enumeration value is specified.
        """
        e = self.assertRaises(ValueError,
            DeliveryInformation.from_element,
            NOTIFICATION_NS.deliveryStatus(
                L.address('tel:27117654321'),
                L.deliveryStatus('WhatIsThis')))
        self.assertEqual(
            "No such delivery status enumeration value: 'WhatIsThis'", str(e))


class SmsNotificationServiceTests(TestCase):
    """
    Tests for `vumi.transports.parlayx.server.SmsNotificationService`.
    """
    def test_process_empty(self):
        """
        `SmsNotificationService.process` raises `SoapFault` if there are no
        actionable child elements in the request body.
        """
        service = SmsNotificationService(None, None)
        exc = self.assertRaises(SoapFault,
            service.process, None, L.root())
        self.assertEqual(
            ('soapenv:Client', 'No actionable items'),
            (exc.code, str(exc)))

    def test_process_unknown(self):
        """
        `SmsNotificationService.process` invokes
        `SmsNotificationService.process_unknown`, for handling otherwise
        unknown requests, which raises `SoapFault`.
        """
        service = SmsNotificationService(None, None)
        exc = self.assertRaises(SoapFault,
            service.process, None, L.root(L.WhatIsThis))
        self.assertEqual(
            ('soapenv:Server', 'No handler for WhatIsThis'),
            (exc.code, str(exc)))

    def test_process_notifySmsReception(self):
        """
        `SmsNotificationService.process_notifySmsReception` invokes the
        message delivery callback with the correlator (message identifier) and
        a `SmsMessage` instance containing the details of the delivered
        message.
        """
        def callback(*a):
            self.callbacks.append(a)
        self.callbacks = []
        service = SmsNotificationService(callback, None)
        self.successResultOf(service.process(None,
            SOAP_ENV.Body(
                create_sms_reception_element(
                    '1234', 'message', '+27117654321', '54321')),
            SOAP_ENV.Header(
                PARLAYX_COMMON_NS.NotifySOAPHeader(
                    PARLAYX_COMMON_NS.linkid('linkid')))))

        self.assertEqual(1, len(self.callbacks))
        correlator, linkid, msg = self.callbacks[0]
        self.assertEqual(
            ('1234', 'linkid', 'message', '+27117654321', '54321', None),
            (correlator, linkid, msg.message, msg.sender_address,
             msg.service_activation_number, msg.timestamp))

    def test_process_notifySmsDeliveryReceipt(self):
        """
        `SmsNotificationService.process_notifySmsDeliveryReceipt` invokes the
        delivery receipt callback with the correlator (message identifier) and
        the delivery status (translated into a Vumi-compatible value.)
        """
        def callback(*a):
            self.callbacks.append(a)
        self.callbacks = []
        service = SmsNotificationService(None, callback)
        self.successResultOf(service.process(None,
            SOAP_ENV.Body(
                create_sms_delivery_receipt(
                    '1234',
                    '+27117654321',
                    DeliveryStatus.DeliveryUncertain))))

        self.assertEqual(1, len(self.callbacks))
        correlator, status = self.callbacks[0]
        self.assertEqual(('1234', 'pending'), self.callbacks[0])

    def test_render(self):
        """
        `SmsNotificationService.render_POST` parses a SOAP request and
        dispatches it to `SmsNotificationService.process` for processing.
        """
        service = SmsNotificationService(None, None)
        service.process = lambda *a, **kw: L.done()
        request = DummyRequest([])
        request.content = StringIO(tostring(soap_envelope('hello')))
        d = request.notifyFinish()
        service.render_POST(request)
        self.successResultOf(d)
        self.assertEqual(http.OK, request.responseCode)
        self.assertEqual(
            {str(SOAP_ENV.Envelope): {
                str(SOAP_ENV.Body): {
                    'done': None}}},
            element_to_dict(fromstring(''.join(request.written))))

    def test_render_soap_fault(self):
        """
        `SmsNotificationService.render_POST` logs any exceptions that occur
        during processing and writes a SOAP fault back to the request. If the
        logged exception is a `SoapFault` its ``to_element`` method is invoked
        to serialize the fault.
        """
        service = SmsNotificationService(None, None)
        service.process = lambda *a, **kw: L.done()
        request = DummyRequest([])
        request.content = StringIO(tostring(L.hello()))
        d = request.notifyFinish()

        service.render_POST(request)
        self.successResultOf(d)
        self.assertEqual(http.INTERNAL_SERVER_ERROR, request.responseCode)
        failures = self.flushLoggedErrors(SoapFault)
        self.assertEqual(1, len(failures))
        self.assertEqual(
            {str(SOAP_ENV.Envelope): {
                str(SOAP_ENV.Body): {
                    str(SOAP_ENV.Fault): {
                        'faultcode': 'soapenv:Client',
                        'faultstring': 'Malformed SOAP request'}}}},
            element_to_dict(fromstring(''.join(request.written))))

    def test_render_exceptions(self):
        """
        `SmsNotificationService.render_POST` logs any exceptions that occur
        during processing and writes a SOAP fault back to the request.
        """
        def process(*a, **kw):
            raise ValueError('What is this')
        service = SmsNotificationService(None, None)
        service.process = process
        request = DummyRequest([])
        request.content = StringIO(tostring(soap_envelope('hello')))
        d = request.notifyFinish()

        service.render_POST(request)
        self.successResultOf(d)
        self.assertEqual(http.INTERNAL_SERVER_ERROR, request.responseCode)
        failures = self.flushLoggedErrors(ValueError)
        self.assertEqual(1, len(failures))
        self.assertEqual(
            {str(SOAP_ENV.Envelope): {
                str(SOAP_ENV.Body): {
                    str(SOAP_ENV.Fault): {
                        'faultcode': 'soapenv:Server',
                        'faultstring': 'What is this'}}}},
            element_to_dict(fromstring(''.join(request.written))))

    def test_render_invalid_xml(self):
        """
        `SmsNotificationService.render_POST` does not accept invalid XML body
        content.
        """
        service = SmsNotificationService(None, None)
        request = DummyRequest([])
        request.content = StringIO('sup')
        d = request.notifyFinish()

        service.render_POST(request)
        self.successResultOf(d)
        self.assertEqual(http.INTERNAL_SERVER_ERROR, request.responseCode)
        failures = self.flushLoggedErrors(ParseError)
        self.assertEqual(1, len(failures))



# -*- test-case-name: vumi.blinkenlights.heartbeat.tests.test_storage -*-

"""
Storage Schema:

 Timestamp (UNIX timestamp):  key = timestamp
 List of systems (JSON list): key = systems
 System state (JSON dict):    key = system:$SYSTEM_ID
 Worker issue (JSON dict):    key = worker:$WORKER_ID:issue
"""

import json

from vumi.persist.redis_base import Manager

TIMESTAMP_KEY = "timestamp"
SYSTEMS_KEY = "systems"


def issue_key(worker_id):
    return "worker:%s:issue" % worker_id


def system_key(system_id):
    return "system:%s" % system_id


class Storage(object):
    """
    TxRedis interface for the heartbeat monitor. Basically only
    supports mutating operations since the monitor does not do
    any reads
    """

    def __init__(self, redis):
        self._redis = redis
        self.manager = redis

    @Manager.calls_manager
    def add_system_ids(self, system_ids):
        yield self._redis.sadd(SYSTEMS_KEY, *system_ids)

    @Manager.calls_manager
    def write_system(self, sys):
        key = system_key(sys.system_id)
        yield self._redis.set(key, sys.dumps())

    def _issue_to_dict(self, issue):
        return {
            'issue_type': issue.issue_type,
            'start_time': issue.start_time,
            'procs_count': issue.procs_count,
        }

    @Manager.calls_manager
    def delete_worker_issue(self, worker_id):
        key = issue_key(worker_id)
        yield self._redis.delete(key)

    @Manager.calls_manager
    def open_or_update_issue(self, worker_id, issue):
        key = issue_key(worker_id)
        issue_raw = yield self._redis.get(key)
        if issue_raw is None:
            issue_data = self._issue_to_dict(issue)
        else:
            issue_data = json.loads(issue_raw)
            issue_data['procs_count'] = issue.procs_count
        yield self._redis.set(key, json.dumps(issue_data))



# -*- test-case-name: vumi.persist.tests.test_fields -*-

"""Field types for Vumi's persistence models."""

import iso8601
from datetime import datetime

from vumi.message import format_vumi_date, parse_vumi_date
from vumi.utils import to_kwargs


# Index values in Riak have to be non-empty, so a zero-length string
# counts as "no value". Since we still have legacy data that was
# inadvertantly indexed with "None" because of the str() call in the
# library and we still have legacy code that relies on an index search
# for a value of "None", fixing this properly here will break existing
# functionality. Once we have rewritten the offending code to not use
# "None" in the index, we can remove the hack below and be happier.
STORE_NONE_FOR_EMPTY_INDEX = False


class ValidationError(Exception):
    """Raised when a value assigned to a field is invalid."""


class FieldDescriptor(object):
    """Property for getting and setting fields."""

    def __init__(self, key, field):
        self.key = key
        self.field = field
        if self.field.index:
            if self.field.index_name is None:
                self.index_name = "%s_bin" % self.key
            else:
                self.index_name = field.index_name
        else:
            self.index_name = None

    def setup(self, model_cls):
        self.model_cls = model_cls

    def validate(self, value):
        self.field.validate(value)

    def initialize(self, modelobj, value):
        self.validate(value)
        self.set_value(modelobj, value)

    def _add_index(self, modelobj, value):
        # XXX: The underlying libraries call str() on whatever index values we
        # provide, so we do this explicitly here and special-case None.
        if value is None:
            if STORE_NONE_FOR_EMPTY_INDEX:
                # Hackery for things that need "None" index values.
                modelobj._riak_object.add_index(self.index_name, "None")
            return
        modelobj._riak_object.add_index(self.index_name, str(value))

    def get_riak_data(self, modelobj, default=None, key=None):
        if key is None:
            key = self.key
        return modelobj._riak_object.get_data().get(key, default)

    def set_riak_data(self, modelobj, raw_value, key=None):
        if key is None:
            key = self.key
        old_raw_value = modelobj._riak_object.get_data().get(key)
        # We set this even if it's "unmodified" because we want explicit `None`
        # values rather than missing fields.
        modelobj._riak_object.set_data_field(key, raw_value)
        if old_raw_value != raw_value:
            modelobj._field_changed(self.key)

    def delete_riak_data(self, modelobj, key=None):
        if key is None:
            key = self.key
        old_raw_value = modelobj._riak_object.get_data().get(key)
        modelobj._riak_object.delete_data_field(key)
        if old_raw_value is not None:
            modelobj._field_changed(self.key)

    def set_value(self, modelobj, value):
        """Set the value associated with this descriptor."""
        raw_value = self.field.to_riak(value)
        self.set_riak_data(modelobj, raw_value)
        if self.index_name is not None:
            modelobj._riak_object.remove_index(self.index_name)
            self._add_index(modelobj, raw_value)

    def get_value(self, modelobj):
        """Get the value associated with this descriptor."""
        return self.field.from_riak(self.get_riak_data(modelobj))

    def clean(self, modelobj):
        """Do any cleanup of the model data for this descriptor after loading
        the data from Riak."""
        pass

    def model_field_changed(self, modelobj, changed_field_name):
        """
        Do any necessary computation when a field changes.
        """
        pass

    def __repr__(self):
        return "<%s key=%s field=%r>" % (self.__class__.__name__, self.key,
                                         self.field)

    def __get__(self, instance, owner):
        if instance is None:
            return self.field
        return self.get_value(instance)

    def __set__(self, instance, value):
        # instance can never be None here
        self.validate(value)
        self.set_value(instance, value)


class Field(object):
    """Base class for model attributes / fields.

    :param object default:
        Default value for the field. The default default is None.
    :param boolean null:
        Whether None is allowed as a value. Default is False (which
        means the field must either be specified explicitly or by
        a non-None default).
    :param boolen index:
        Whether the field should also be indexed. Default is False.
    :param string index_name:
        The name to use for the index. The default is the field name
        followed by _bin.
    """

    descriptor_class = FieldDescriptor
    # whether an attempt should be made to initialize the field on
    # model instance creation
    initializable = True

    def __init__(self, default=None, null=False, index=False, index_name=None):
        self.default = default
        self.null = null
        self.index = index
        self.index_name = index_name

    def get_descriptor(self, key):
        return self.descriptor_class(key, self)

    def validate(self, value):
        """Validate a value.

        Checks null values and calls .validate() for non-null
        values. Raises ValidationError if a value is invalid.
        """
        if not self.null and value is None:
            raise ValidationError("None is not allowed as a value for non-null"
                                  " fields.")
        if value is not None:
            self.custom_validate(value)

    def custom_validate(self, value):
        """Check whether a non-null value is valid for this field."""
        pass

    def to_riak(self, value):
        return self.custom_to_riak(value) if value is not None else None

    def custom_to_riak(self, value):
        """Convert a non-None value to something storable by Riak."""
        return value

    def from_riak(self, raw_value):
        return (self.custom_from_riak(raw_value)
                if raw_value is not None else None)

    def custom_from_riak(self, raw_value):
        """Convert a non-None value stored by Riak to Python."""
        return raw_value


class Integer(Field):
    """Field that accepts integers.

    Additional parameters:

    :param integer min:
        Minimum allowed value (default is `None` which indicates no minimum).
    :param integer max:
        Maximum allowed value (default is `None` which indicates no maximum).
    """
    def __init__(self, min=None, max=None, **kw):
        super(Integer, self).__init__(**kw)
        self.min = min
        self.max = max

    def custom_validate(self, value):
        if not isinstance(value, (int, long)):
            raise ValidationError("Value %r is not an integer." % (value,))
        if self.min is not None and value < self.min:
            raise ValidationError("Value %r too low (minimum value is %d)."
                                  % (value, self.min))
        if self.max is not None and value > self.max:
            raise ValidationError("Value %r too high (maximum value is %d)."
                                  % (value, self.max))


class Boolean(Field):
    """Field that is either True or False.
    """

    def custom_validate(self, value):
        if not isinstance(value, bool):
            raise ValidationError('Value %r is not a boolean.' % (value,))


class Unicode(Field):
    """Field that accepts unicode strings.

    Additional parameters:

    :param integer max_length:
        Maximum allowed length (default is `None` which indicates no maximum).
    """
    def __init__(self, max_length=None, **kw):
        super(Unicode, self).__init__(**kw)
        self.max_length = max_length

    def custom_validate(self, value):
        if not isinstance(value, unicode):
            raise ValidationError("Value %r is not a unicode string."
                                  % (value,))
        if self.max_length is not None and len(value) > self.max_length:
            raise ValidationError("Value %r too long (maximum length is %d)."
                                  % (value, self.max_length))


class Tag(Field):
    """Field that represents a Vumi tag."""
    def custom_validate(self, value):
        if not isinstance(value, tuple) or len(value) != 2:
            raise ValidationError("Tags %r should be a (pool, tag_name)"
                                  " tuple" % (value,))

    def custom_to_riak(self, value):
        return list(value)

    def custom_from_riak(self, value):
        return tuple(value)


class TimestampDescriptor(FieldDescriptor):
    """A field descriptor for timestamp fields."""

    def set_value(self, modelobj, value):
        if value is not None and not isinstance(value, datetime):
            # we can be sure that this is a iso8601 parseable string, since it
            # passed validation
            value = iso8601.parse_date(value)
        super(TimestampDescriptor, self).set_value(modelobj, value)


class Timestamp(Field):
    """Field that stores a datetime."""

    descriptor_class = TimestampDescriptor

    def custom_validate(self, value):
        if isinstance(value, datetime):
            return

        try:
            iso8601.parse_date(value)
            return
        except iso8601.ParseError:
            pass

        raise ValidationError("Timestamp field expects a datetime or an "
                              "iso8601 formatted string.")

    def custom_to_riak(self, value):
        return format_vumi_date(value)

    def custom_from_riak(self, value):
        return parse_vumi_date(value)


class Json(Field):
    """Field that stores an object that can be serialized to/from JSON."""
    pass


class VumiMessageDescriptor(FieldDescriptor):
    """Property for getting and setting fields."""

    def setup(self, model_cls):
        super(VumiMessageDescriptor, self).setup(model_cls)
        self.message_class = self.field.message_class
        if self.field.prefix is None:
            self.prefix = "%s." % self.key
        else:
            self.prefix = self.field.prefix

    def _clear_keys(self, modelobj):
        for key in modelobj._riak_object.get_data().keys():
            if key.startswith(self.prefix):
                self.delete_riak_data(modelobj, key)

    def _timestamp_to_json(self, dt):
        return format_vumi_date(dt)

    def _timestamp_from_json(self, value):
        return parse_vumi_date(value)

    def set_value(self, modelobj, msg):
        """Set the value associated with this descriptor."""
        self._clear_keys(modelobj)
        if msg is None:
            return
        for key, value in msg.payload.iteritems():
            if key == self.message_class._CACHE_ATTRIBUTE:
                continue
            # TODO: timestamp as datetime in payload must die.
            if key == "timestamp":
                value = self._timestamp_to_json(value)
            full_key = "%s%s" % (self.prefix, key)
            self.set_riak_data(modelobj, value, full_key)

    def get_value(self, modelobj):
        """Get the value associated with this descriptor."""
        payload = {}
        for key, value in modelobj._riak_object.get_data().iteritems():
            if key.startswith(self.prefix):
                key = key[len(self.prefix):]
                # TODO: timestamp as datetime in payload must die.
                if key == "timestamp":
                    value = self._timestamp_from_json(value)
                payload[key] = value
        if not payload:
            return None
        return self.field.message_class(**to_kwargs(payload))


class VumiMessage(Field):
    """Field that represents a Vumi message.

    Additional parameters:

    :param class message_class:
        The class of the message objects being stored.
        Usually one of Message, TransportUserMessage or TransportEvent.
    :param string prefix:
        The prefix to use when storing message payload keys in Riak. Default is
        the name of the field followed by a dot ('.').

    Note::

       The special message attribute ``__cache__`` is not stored by this
       field.
    """
    descriptor_class = VumiMessageDescriptor

    def __init__(self, message_class, prefix=None, **kw):
        super(VumiMessage, self).__init__(**kw)
        self.message_class = message_class
        self.prefix = prefix

    def custom_validate(self, value):
        if not isinstance(value, self.message_class):
            raise ValidationError("Message %r should be an instance of %r"
                                  % (value, self.message_class))


class FieldWithSubtype(Field):
    """Base class for a field that is a collection of other fields of a
    single type.

    :param Field field_type:
        The field specification for the dynamic values. Default is Unicode().
    """
    def __init__(self, field_type=None, **kw):
        super(FieldWithSubtype, self).__init__(**kw)
        if field_type is None:
            field_type = Unicode()
        if field_type.descriptor_class is not FieldDescriptor:
            raise RuntimeError("Dynamic fields only supports fields that"
                               " that use the basic FieldDescriptor class")
        self.field_type = field_type

    def validate_subfield(self, value):
        self.field_type.validate(value)

    def subfield_to_riak(self, value):
        return self.field_type.to_riak(value)

    def subfield_from_riak(self, value):
        return self.field_type.from_riak(value)


class DynamicDescriptor(FieldDescriptor):
    """A field descriptor for dynamic fields."""
    def setup(self, model_cls):
        super(DynamicDescriptor, self).setup(model_cls)
        if self.field.prefix is None:
            self.prefix = "%s." % self.key
        else:
            self.prefix = self.field.prefix

    def initialize(self, modelobj, valuedict):
        if valuedict is not None:
            self.update(modelobj, valuedict)

    def get_value(self, modelobj):
        return DynamicProxy(self, modelobj)

    def set_value(self, modelobj, valuedict):
        self.clear(modelobj)
        self.update(modelobj, valuedict)

    def clear(self, modelobj):
        keys = list(self.iterkeys(modelobj))
        for key in keys:
            self.delete_dynamic_value(modelobj, key)

    def iterkeys(self, modelobj):
        prefix_len = len(self.prefix)
        data = modelobj._riak_object.get_data()
        return (key[prefix_len:]
                for key in data.iterkeys()
                if key.startswith(self.prefix))

    def iteritems(self, modelobj):
        prefix_len = len(self.prefix)
        from_riak = self.field.subfield_from_riak
        data = modelobj._riak_object.get_data()
        return ((key[prefix_len:], from_riak(value))
                for key, value in data.iteritems()
                if key.startswith(self.prefix))

    def update(self, modelobj, otherdict):
        # this is a separate method so it can succeed or fail
        # somewhat atomically in the case where otherdict contains
        # bad keys or values
        items = [(self.prefix + key, self.field.subfield_to_riak(value))
                 for key, value in otherdict.iteritems()]
        for key, value in items:
            self.set_riak_data(modelobj, value, key=key)

    def get_dynamic_value(self, modelobj, dynamic_key):
        return self.field.subfield_from_riak(
            self.get_riak_data(modelobj, key=(self.prefix + dynamic_key)))

    def set_dynamic_value(self, modelobj, dynamic_key, value):
        self.field.validate_subfield(value)
        value = self.field.subfield_to_riak(value)
        self.set_riak_data(modelobj, value, self.prefix + dynamic_key)

    def delete_dynamic_value(self, modelobj, dynamic_key):
        self.delete_riak_data(modelobj, self.prefix + dynamic_key)

    def has_dynamic_key(self, modelobj, dynamic_key):
        key = self.prefix + dynamic_key
        return key in modelobj._riak_object.get_data()


class DynamicProxy(object):
    def __init__(self, descriptor, modelobj):
        self._descriptor = descriptor
        self._modelobj = modelobj

    def iterkeys(self):
        return self._descriptor.iterkeys(self._modelobj)

    def keys(self):
        return list(self.iterkeys())

    def iteritems(self):
        return self._descriptor.iteritems(self._modelobj)

    def items(self):
        return list(self.iteritems())

    def itervalues(self):
        return (value for _key, value in self.iteritems())

    def values(self):
        return list(self.itervalues())

    def update(self, otherdict):
        return self._descriptor.update(self._modelobj, otherdict)

    def clear(self):
        self._descriptor.clear(self._modelobj)

    def copy(self):
        return dict(self.iteritems())

    def __getitem__(self, key):
        return self._descriptor.get_dynamic_value(self._modelobj, key)

    def __setitem__(self, key, value):
        self._descriptor.set_dynamic_value(self._modelobj, key, value)

    def __delitem__(self, key):
        self._descriptor.delete_dynamic_value(self._modelobj, key)

    def __contains__(self, key):
        return self._descriptor.has_dynamic_key(self._modelobj, key)


class Dynamic(FieldWithSubtype):
    """A field that allows sub-fields to be added dynamically.

    :param Field field_type:
        The field specification for the dynamic values. Default is Unicode().
    :param string prefix:
        The prefix to use when storing these values in Riak. Default is
        the name of the field followed by a dot ('.').
    """
    descriptor_class = DynamicDescriptor

    def __init__(self, field_type=None, prefix=None):
        super(Dynamic, self).__init__(field_type=field_type)
        self.prefix = prefix

    def custom_validate(self, valuedict):
        if not isinstance(valuedict, dict):
            raise ValidationError(
                "Value %r should be a dict of subfield name-value pairs"
                % valuedict)
        for key, value in valuedict.iteritems():
            self.validate_subfield(value)
            if not isinstance(key, unicode):
                raise ValidationError("Dynamic field needs unicode keys.")


class ListOfDescriptor(FieldDescriptor):
    """A field descriptor for ListOf fields."""

    def get_value(self, modelobj):
        return ListProxy(self, modelobj)

    def get_list_item(self, modelobj, list_idx):
        raw_item = self.get_riak_data(modelobj, [])[list_idx]
        return self.field.subfield_from_riak(raw_item)

    def _set_model_data(self, modelobj, raw_values):
        self.set_riak_data(modelobj, raw_values)
        if self.index_name is not None:
            modelobj._riak_object.remove_index(self.index_name)
            for value in raw_values:
                self._add_index(modelobj, value)

    def set_value(self, modelobj, values):
        map(self.field.validate_subfield, values)
        raw_values = [self.field.subfield_to_riak(value) for value in values]
        self._set_model_data(modelobj, raw_values)

    def set_list_item(self, modelobj, list_idx, value):
        self.field.validate_subfield(value)
        raw_value = self.field.subfield_to_riak(value)
        field_list = self.get_riak_data(modelobj, [])
        field_list[list_idx] = raw_value
        self._set_model_data(modelobj, field_list)

    def del_list_item(self, modelobj, list_idx):
        field_list = self.get_riak_data(modelobj, [])
        del field_list[list_idx]
        self._set_model_data(modelobj, field_list)

    def append_list_item(self, modelobj, value):
        self.field.validate_subfield(value)
        raw_value = self.field.subfield_to_riak(value)
        field_list = self.get_riak_data(modelobj, [])
        field_list.append(raw_value)
        self._set_model_data(modelobj, field_list)

    def remove_list_item(self, modelobj, value):
        self.field.validate_subfield(value)
        raw_value = self.field.subfield_to_riak(value)
        field_list = self.get_riak_data(modelobj, [])
        field_list.remove(raw_value)
        self._set_model_data(modelobj, field_list)

    def extend_list(self, modelobj, values):
        map(self.field.validate_subfield, values)
        raw_values = [self.field.subfield_to_riak(value) for value in values]
        field_list = self.get_riak_data(modelobj, [])
        field_list.extend(raw_values)
        self._set_model_data(modelobj, field_list)

    def iter_list(self, modelobj):
        raw_list = self.get_riak_data(modelobj, [])
        for raw_value in raw_list:
            yield self.field.subfield_from_riak(raw_value)


class ListProxy(object):
    def __init__(self, descriptor, modelobj):
        self._descriptor = descriptor
        self._modelobj = modelobj

    def __getitem__(self, idx):
        return self._descriptor.get_list_item(self._modelobj, idx)

    def __setitem__(self, idx, value):
        self._descriptor.set_list_item(self._modelobj, idx, value)

    def __delitem__(self, idx):
        self._descriptor.del_list_item(self._modelobj, idx)

    def remove(self, value):
        self._descriptor.remove_list_item(self._modelobj, value)

    def append(self, value):
        self._descriptor.append_list_item(self._modelobj, value)

    def extend(self, values):
        self._descriptor.extend_list(self._modelobj, values)

    def __iter__(self):
        return self._descriptor.iter_list(self._modelobj)


class ListOf(FieldWithSubtype):
    """A field that contains a list of values of some other type.

    :param Field field_type:
    The field specification for the dynamic values. Default is Unicode().
    """
    descriptor_class = ListOfDescriptor

    def __init__(self, field_type=None, **kw):
        super(ListOf, self).__init__(field_type=field_type, default=list, **kw)

    def custom_validate(self, valuelist):
        if not isinstance(valuelist, list):
            raise ValidationError(
                "Value %r should be a list of values" % valuelist)
        map(self.validate_subfield, valuelist)


class SetOfDescriptor(FieldDescriptor):
    """
    A field descriptor for SetOf fields.
    """

    def get_value(self, modelobj):
        return SetProxy(self, modelobj)

    def _get_model_data(self, modelobj):
        return set(self.get_riak_data(modelobj, []))

    def _set_model_data(self, modelobj, raw_values):
        raw_values = sorted(set(raw_values))
        self.set_riak_data(modelobj, raw_values)
        if self.index_name is not None:
            modelobj._riak_object.remove_index(self.index_name)
            for value in raw_values:
                self._add_index(modelobj, value)

    def set_contains_item(self, modelobj, value):
        field_set = self._get_model_data(modelobj)
        return value in field_set

    def set_value(self, modelobj, values):
        map(self.field.validate_subfield, values)
        raw_values = [self.field.subfield_to_riak(value) for value in values]
        self._set_model_data(modelobj, raw_values)

    def add_set_item(self, modelobj, value):
        self.field.validate_subfield(value)
        field_set = self._get_model_data(modelobj)
        field_set.add(self.field.subfield_to_riak(value))
        self._set_model_data(modelobj, field_set)

    def remove_set_item(self, modelobj, value):
        self.field.validate_subfield(value)
        field_set = self._get_model_data(modelobj)
        field_set.remove(value)
        self._set_model_data(modelobj, field_set)

    def discard_set_item(self, modelobj, value):
        self.field.validate_subfield(value)
        field_set = self._get_model_data(modelobj)
        field_set.discard(value)
        self._set_model_data(modelobj, field_set)

    def update_set(self, modelobj, values):
        map(self.field.validate_subfield, values)
        raw_values = [self.field.subfield_to_riak(value) for value in values]
        field_set = self._get_model_data(modelobj)
        field_set.update(raw_values)
        self._set_model_data(modelobj, field_set)

    def iter_set(self, modelobj):
        field_set = self._get_model_data(modelobj)
        for raw_value in field_set:
            yield self.field.subfield_from_riak(raw_value)


class SetProxy(object):
    def __init__(self, descriptor, modelobj):
        self._descriptor = descriptor
        self._modelobj = modelobj

    def __contains__(self, value):
        return self._descriptor.set_contains_item(self._modelobj, value)

    def add(self, value):
        self._descriptor.add_set_item(self._modelobj, value)

    def remove(self, value):
        self._descriptor.remove_set_item(self._modelobj, value)

    def discard(self, value):
        self._descriptor.discard_set_item(self._modelobj, value)

    def update(self, values):
        self._descriptor.update_set(self._modelobj, values)

    def __iter__(self):
        return self._descriptor.iter_set(self._modelobj)


class SetOf(FieldWithSubtype):
    """
    A field that contains a set of values of some other type.

    :param Field field_type:
        The field specification for the dynamic values. Default is Unicode().
    """
    descriptor_class = SetOfDescriptor

    def __init__(self, field_type=None, **kw):
        super(SetOf, self).__init__(field_type=field_type, default=set, **kw)

    def custom_validate(self, valueset):
        if not isinstance(valueset, set):
            raise ValidationError(
                "Value %r should be a set of values" % valueset)
        map(self.validate_subfield, valueset)

    def custom_to_riak(self, value):
        return sorted(value)

    def custom_from_riak(self, raw_value):
        return set(raw_value)


class ForeignKeyDescriptor(FieldDescriptor):
    def setup(self, model_cls):
        super(ForeignKeyDescriptor, self).setup(model_cls)
        self.other_model = self.field.other_model
        if self.field.index is None:
            self.index_name = "%s_bin" % self.key
        else:
            self.index_name = self.field.index

        backlink_name = self.field.backlink
        if backlink_name is None:
            backlink_name = model_cls.__name__.lower() + "s"
        self.other_model.backlinks.declare_backlink(
            backlink_name, self.reverse_lookup_keys)

        backlink_keys_name = backlink_name + "_keys"
        if backlink_keys_name.endswith("s_keys"):
            backlink_keys_name = backlink_name[:-1] + "_keys"
        self.other_model.backlinks.declare_backlink(
            backlink_keys_name, self.reverse_lookup_keys_paginated)

    def reverse_lookup_keys(self, modelobj, manager=None):
        if manager is None:
            manager = modelobj.manager
        return manager.index_keys(
            self.model_cls, self.index_name, modelobj.key)

    def reverse_lookup_keys_paginated(self, modelobj, manager=None,
                                      max_results=None, continuation=None):
        """
        Perform a paginated index query for backlinked objects.
        """
        if manager is None:
            manager = modelobj.manager
        return manager.index_keys_page(
            self.model_cls, self.index_name, modelobj.key,
            max_results=max_results, continuation=continuation)

    def clean(self, modelobj):
        if self.key not in modelobj._riak_object.get_data():
            # We might have an old-style index-only version of the data.
            indexes = [
                value for name, value in modelobj._riak_object.get_indexes()
                if name == self.index_name]
            self.set_riak_data(modelobj, (indexes or [None])[0])

    def get_value(self, modelobj):
        return ForeignKeyProxy(self, modelobj)

    def get_foreign_key(self, modelobj):
        return self.get_riak_data(modelobj)

    def set_foreign_key(self, modelobj, foreign_key):
        self.set_riak_data(modelobj, foreign_key)
        modelobj._riak_object.remove_index(self.index_name)
        if foreign_key is not None:
            self._add_index(modelobj, foreign_key)

    def get_foreign_object(self, modelobj, manager=None):
        key = self.get_foreign_key(modelobj)
        if key is None:
            return None
        if manager is None:
            manager = modelobj.manager
        return self.other_model.load(manager, key)

    def initialize(self, modelobj, value):
        if isinstance(value, basestring):
            self.set_foreign_key(modelobj, value)
        else:
            self.set_foreign_object(modelobj, value)

    def set_value(self, modelobj, value):
        raise RuntimeError("ForeignKeyDescriptors should never be assigned"
                           " to.")

    def set_foreign_object(self, modelobj, otherobj):
        self.validate(otherobj)
        foreign_key = otherobj.key if otherobj is not None else None
        self.set_foreign_key(modelobj, foreign_key)


class ForeignKeyProxy(object):
    def __init__(self, descriptor, modelobj):
        self._descriptor = descriptor
        self._modelobj = modelobj

    def _get_key(self):
        return self._descriptor.get_foreign_key(self._modelobj)

    def _set_key(self, foreign_key):
        return self._descriptor.set_foreign_key(self._modelobj, foreign_key)

    key = property(fget=_get_key, fset=_set_key)

    def get(self, manager=None):
        return self._descriptor.get_foreign_object(self._modelobj, manager)

    def set(self, otherobj):
        self._descriptor.set_foreign_object(self._modelobj, otherobj)


class ForeignKey(Field):
    """A field that links to another class.

    Additional parameters:

    :param Model other_model:
        The type of model linked to.
    :param string index:
        The name to use for the index. The default is the field name
        followed by _bin.
    :param string backlink:
        The name to use for the backlink on :attr:`other_model.backlinks`.
        The default is the name of the class the field is on converted
        to lowercase and with 's' appended (e.g. 'FooModel' would result
        in :attr:`other_model.backlinks.foomodels`). This is also used (with
        `_keys` appended and a trailing `s` omitted if one is present) for the
        paginated keys backlink function.
    """
    descriptor_class = ForeignKeyDescriptor

    def __init__(self, other_model, index=None, backlink=None, **kw):
        super(ForeignKey, self).__init__(**kw)
        self.other_model = other_model
        self.index = index
        self.backlink = backlink

    def custom_validate(self, value):
        if not isinstance(value, self.other_model):
            raise ValidationError("ForeignKey requires a %r instance" %
                                  (self.other_model,))


class ManyToManyDescriptor(ForeignKeyDescriptor):

    def get_value(self, modelobj):
        return ManyToManyProxy(self, modelobj)

    def set_value(self, modelobj, value):
        raise RuntimeError("ManyToManyDescriptors should never be assigned"
                           " to.")

    def clean(self, modelobj):
        if self.key not in modelobj._riak_object.get_data():
            # We might have an old-style index-only version of the data.
            indexes = [
                value for name, value in modelobj._riak_object.get_indexes()
                if name == self.index_name]
            self.set_riak_data(modelobj, indexes[:])

    def get_foreign_keys(self, modelobj):
        return self.get_riak_data(modelobj, [])[:]

    def add_foreign_key(self, modelobj, foreign_key):
        if foreign_key not in self.get_foreign_keys(modelobj):
            field_list = self.get_riak_data(modelobj, [])
            field_list.append(foreign_key)
            self.set_riak_data(modelobj, field_list)
        self._add_index(modelobj, foreign_key)

    def remove_foreign_key(self, modelobj, foreign_key):
        if foreign_key in self.get_foreign_keys(modelobj):
            field_list = self.get_riak_data(modelobj, [])
            field_list.remove(foreign_key)
            self.set_riak_data(modelobj, field_list)
        modelobj._riak_object.remove_index(self.index_name, foreign_key)

    def load_foreign_objects(self, modelobj, manager=None):
        keys = self.get_foreign_keys(modelobj)
        if manager is None:
            manager = modelobj.manager
        return manager.load_all_bunches(self.other_model, keys)

    def add_foreign_object(self, modelobj, otherobj):
        self.validate(otherobj)
        self.add_foreign_key(modelobj, otherobj.key)

    def remove_foreign_object(self, modelobj, otherobj):
        self.validate(otherobj)
        self.remove_foreign_key(modelobj, otherobj.key)

    def clear_keys(self, modelobj):
        self.set_riak_data(modelobj, [])
        modelobj._riak_object.remove_index(self.index_name)


class ManyToManyProxy(object):
    def __init__(self, descriptor, modelobj):
        self._descriptor = descriptor
        self._modelobj = modelobj

    def keys(self):
        return self._descriptor.get_foreign_keys(self._modelobj)

    def add_key(self, foreign_key):
        self._descriptor.add_foreign_key(self._modelobj, foreign_key)

    def remove_key(self, foreign_key):
        self._descriptor.remove_foreign_key(self._modelobj, foreign_key)

    def load_all_bunches(self, manager=None):
        return self._descriptor.load_foreign_objects(self._modelobj, manager)

    def add(self, otherobj):
        self._descriptor.add_foreign_object(self._modelobj, otherobj)

    def remove(self, otherobj):
        self._descriptor.remove_foreign_object(self._modelobj, otherobj)

    def clear(self):
        self._descriptor.clear_keys(self._modelobj)


class ManyToMany(ForeignKey):
    """A field that links to multiple instances of another class.

    :param Model other_model:
        The type of model linked to.
    :param string index:
        The name to use for the index. The default is the field name
        followed by _bin.
    :param string backlink:
        The name to use for the backlink on :attr:`other_model.backlinks`.
        The default is the name of the class the field is on converted
        to lowercase and with 's' appended (e.g. 'FooModel' would result
        in :attr:`other_model.backlinks.foomodels`). This is also used (with
        `_keys` appended and a trailing `s` omitted if one is present) for the
        paginated keys backlink function.
    """
    descriptor_class = ManyToManyDescriptor
    initializable = False

    def __init__(self, other_model, index=None, backlink=None):
        super(ManyToMany, self).__init__(other_model, index, backlink)


class ComputedValueDescriptor(FieldDescriptor):
    """A field descriptor for computed value fields."""

    def __init__(self, key, field):
        super(ComputedValueDescriptor, self).__init__(key, field)
        self.subfield_descriptor = field.field_type.get_descriptor(key)

    def model_field_changed(self, modelobj, changed_field_name):
        self.set_value(modelobj, self.field.value_func(modelobj))

    def set_value(self, modelobj, value):
        return self.subfield_descriptor.set_value(modelobj, value)

    def get_value(self, modelobj):
        return self.subfield_descriptor.get_value(modelobj)

    def __set__(self, instance, value):
        raise RuntimeError("Can't set value of computed field.")


class ComputedValue(Field):
    """
    Field that stores a computed value.

    :param value_func:
        A function that takes a model instance as its only parameter and
        returns a value for the field. This is called whenever the value of
        another field changes to compute the value of this field.

    :param Field field_type:
        The field specification for the computed value. Default is Unicode().
    """

    descriptor_class = ComputedValueDescriptor
    initializable = False

    def __init__(self, value_func, field_type=None):
        super(ComputedValue, self).__init__()
        if field_type is None:
            field_type = Unicode()
        if not isinstance(field_type, Field):
            raise TypeError("field_type must be a Field object.")
        self.field_type = field_type
        self.value_func = value_func

    def validate(self, value):
        return self.field_type.validate(value)

    def to_riak(self, value):
        return self.field_type.to_riak(value)

    def from_riak(self, value):
        return self.field_type.from_riak(value)
